/* addrslv.c --- Address resolver implementation

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "addrslv.h"
#include <glib.h>
#include <string.h>

/* Cache to get bfddata from filename */
static GData * addrslv_bfds = NULL;

/* Cache to get symbol info from address */
typedef struct _AddrslvBFDData AddrslvBFDData;
struct _AddrslvBFDData
{
  GTree*   symbol_infos;
  GQuark   libfile_quark;
  guint    virt_addr;
  guint    ref_count;
};

typedef struct _AddrslvPMapsAttachedData AddrslvPMapsAttachedData;
struct _AddrslvPMapsAttachedData
{
  PMaps*   pmaps;
  gboolean hooked_to_update;
  gulong   hook_id;
  GSList*  gc_marked_bfddata;
};

typedef struct _AddrslvBfddataFindData AddrslvBfddataFindData;
struct _AddrslvBfddataFindData
{
  guint  address;
  GTree *tree;
};

/* Address resolver */
static void               addrslv_lock_pmaps       (PMaps * pmaps, 
						    gboolean after, 
						    gpointer callback_data);
static void               addrslv_lock_pmaps_entry_before (PMapsEntry * entry, 
							   gpointer callback_data);
static void               addrslv_lock_pmaps_entry_after (PMapsEntry * entry, 
							  gpointer callback_data);

/* Symbol Info */
static AddrslvSymbolInfo* addrslv_symbolinfo_new  (const gchar * funcname,
						   guint address,
						   guint size,
						   gboolean dynamic,
						   const gchar * filename,
						   guint lineno);
static void               addrslv_symbolinfo_free (AddrslvSymbolInfo * symbol_info);

static gboolean           addrslv_symbolinfo_search_by_funcname (gpointer key,
								 gpointer value,
								 gpointer data);
/* BFD Data */
#define BFDDATA_KEY "addrslv_bfddata" 
static AddrslvBFDData*    addrslv_bfddata_get     (const gchar * filename, GError **err);
static AddrslvBFDData*    addrslv_bfddata_try_get (const gchar * filename);
static AddrslvBFDData*    addrslv_bfddata_new     (const gchar * filename, GError **err);
static void               addrslv_bfddata_free    (AddrslvBFDData * bfddata);
static AddrslvBFDData*    addrslv_bfddata_ref     (AddrslvBFDData * bfddata);
static void               addrslv_bfddata_unref   (AddrslvBFDData * bfddata);
static AddrslvSymbolInfo* addrslv_bfddata_find    (AddrslvBFDData * bfddata, 
						   guint address);

static gint               addrslv_bfddata_find_compare (gconstpointer a,
							gconstpointer b);

static guint              addrslv_bfddata_run_nm          (AddrslvBFDData* bfddata, 
							   const gchar* filename,  
							   gboolean dynamic, 
							   GError **err);

static guint              addrslv_bfddata_run_readelf     (const gchar* filename,  
							   GError **err);

static gboolean           addrslv_bfddata_install_symbol  (AddrslvBFDData* bfddata, 
							   const gchar* buffer,
							   gboolean dynamic);
static gint               addrslv_bfddata_address_compare (gconstpointer a,
							   gconstpointer b, 
							   gpointer user_data);

/* PAttach: Data Attached to PMaps */
#define PATTACH_KEY "addrslv_pattach" 
static AddrslvPMapsAttachedData* addrslv_pattach_get     (PMaps * pmaps);
static AddrslvPMapsAttachedData* addrslv_pattach_try_get (PMaps * pmaps);
static AddrslvPMapsAttachedData* addrslv_pattach_new     (PMaps * pmaps);
static void                      addrslv_pattach_free    (AddrslvPMapsAttachedData* pattach);
static void                      addrslv_pattach_gc_mark (AddrslvPMapsAttachedData* pattach,
							  AddrslvBFDData* bfddata);
static void                      addrslv_pattach_gc_sweep (AddrslvPMapsAttachedData* pattach);

/* -------------------------------------------------------------------
   Address resolver
   ----------------------------------------------------------------- */ 
void
addrslv_init             (void)
{
  if (addrslv_bfds == NULL)
    g_datalist_init (&addrslv_bfds);
}

AddrslvSymbolInfo *
addrslv_resolve(PMaps * pmaps, guint address, GError ** err)
{
  PMapsEntry * pmaps_entry;
  AddrslvBFDData * bfddata;
  GError * tmp_error = NULL;
  
  g_return_val_if_fail (pmaps, NULL);
  g_return_val_if_fail (address, NULL);

  pmaps_entry = pmaps_find(pmaps, address);
  if ( pmaps_entry == NULL )
    return NULL;
    
  if (!(pmaps_entry->prot & PMAPS_EXECUTABLE ))
    return NULL;

  bfddata = pmaps_entry_get_data(pmaps_entry, BFDDATA_KEY);
  if (bfddata == NULL)
    {
      AddrslvPMapsAttachedData * pattach;
      bfddata = addrslv_bfddata_get(pmaps_entry->filename, 
				    &tmp_error);
      if (tmp_error != NULL)
	{
	  g_propagate_error (err, tmp_error);
	  return NULL;
	}
      else if (bfddata == NULL)
	return NULL;
      
      pmaps_entry_set_data(pmaps_entry,
			   BFDDATA_KEY,
			   bfddata, 
			   (GDestroyNotify)addrslv_bfddata_unref);
      
      pattach = addrslv_pattach_get(pmaps);
      if (pattach->hooked_to_update == FALSE)
	{
	  pattach->hook_id = pmaps_add_update_notify(pmaps, 
						     addrslv_lock_pmaps, 
						     NULL);
	  pattach->hooked_to_update = TRUE;
	}
    }

  {
    AddrslvSymbolInfo * symbol_info;
    guint resolve_offset;

    resolve_offset = (pmaps_entry->start - bfddata->virt_addr);
    symbol_info = addrslv_bfddata_find(bfddata, address - resolve_offset);
    if (symbol_info)
      return symbol_info;

    /* Next line may be unnecessary. */
    return addrslv_bfddata_find(bfddata, address);
  }
}

AddrslvSymbolInfo*
addrslv_reverse_resolve  (PMaps * pmaps, 
			  PMapsEntry * entry,
			  const gchar * funcname,
			  GError ** err)
{
  GError* tmp_error 	   = NULL;
  AddrslvBFDData * bfddata;
  GTree* symbol_infos;
  gpointer data;
  
  g_return_val_if_fail (pmaps, NULL);
  g_return_val_if_fail (entry, NULL);
  g_return_val_if_fail (funcname, NULL);

  addrslv_resolve(pmaps, entry->start, &tmp_error);
  if (tmp_error != NULL)
    {
      g_propagate_error (err, tmp_error);
      return NULL;
    }

  bfddata = pmaps_entry_get_data(entry, BFDDATA_KEY);
  if (bfddata == NULL)
    return NULL;

  symbol_infos = bfddata->symbol_infos;
  data 	       = (gchar *)funcname;
  g_tree_foreach(symbol_infos, addrslv_symbolinfo_search_by_funcname, &data);

  if (data == funcname)
    return NULL;
  else
    return data;
}



void
addrslv_fin (void)
{
  g_datalist_clear(&addrslv_bfds);
}

/* addrslv_lock_pmaps
   ------------------
  Mark and sweep garbage collection
  Before updating pmaps, all user_data(AddrslvBFDData) assoicated with pmaps 
  entries in PMAPS are ref'ed.
  After updating pmaps, useful AddrslvBFDData are kept; the other are
  unerf'ed. */
static void
addrslv_lock_pmaps(PMaps * pmaps, gboolean after, gpointer unused)
{
  AddrslvPMapsAttachedData * pattach;
  g_return_if_fail (pmaps);

  if (after)
    {				
      /* sweep phase */
      
      /* Mark all reusable data */
      pmaps_foreach(pmaps, addrslv_lock_pmaps_entry_after, pmaps);

      /* Sweep the garbages */
      pattach = addrslv_pattach_try_get(pmaps);
      g_assert(pattach);
      addrslv_pattach_gc_sweep(pattach);
    }
  else
    {
      pattach = addrslv_pattach_get(pmaps);
      g_assert(pattach->gc_marked_bfddata == NULL);
      
      /* mark phase */
      pmaps_foreach(pmaps, addrslv_lock_pmaps_entry_before, pattach);
    }
}

static void
addrslv_lock_pmaps_entry_before (PMapsEntry * entry, gpointer callback_data)
{
  AddrslvPMapsAttachedData * pattach = callback_data;
  AddrslvBFDData* bfddata;
  g_return_if_fail (entry);
  g_return_if_fail (pattach);
  
  bfddata = pmaps_entry_get_data(entry, BFDDATA_KEY);
  if (bfddata == NULL)
    return ;

  addrslv_pattach_gc_mark(pattach, bfddata);
}

static void
addrslv_lock_pmaps_entry_after (PMapsEntry * entry, gpointer callback_data)
{
  AddrslvBFDData* bfddata;

  bfddata = addrslv_bfddata_try_get(entry->filename);
  if (bfddata == NULL)
    return ;

  addrslv_bfddata_ref(bfddata);
  pmaps_entry_set_data(entry, 
		       BFDDATA_KEY,
		       bfddata, 
		       (GDestroyNotify)addrslv_bfddata_unref);
}

/* -------------------------------------------------------------------
   BFD Data
   ----------------------------------------------------------------- */ 
static AddrslvBFDData*
addrslv_bfddata_get(const gchar* filename, GError** err)
{
  AddrslvBFDData* bfddata = NULL;
  GError * tmp_error = NULL;
  
  bfddata = addrslv_bfddata_try_get(filename);
  if (bfddata)
    return bfddata;
  
  bfddata = addrslv_bfddata_new(filename, &tmp_error);
  if (bfddata == NULL)
    {
      g_propagate_error (err, tmp_error);
      return NULL;
    }
  
  /* addrslv_bfds is not primary owner of bfddata.
     ref_count is not touched. */
  g_datalist_set_data(&addrslv_bfds, filename, bfddata);

  return bfddata;
}

static AddrslvBFDData*
addrslv_bfddata_try_get(const gchar * filename)
{
  g_return_val_if_fail (filename, NULL);

  return g_datalist_get_data(&addrslv_bfds, filename);
}

static AddrslvBFDData*
addrslv_bfddata_new(const gchar * filename, GError **err)
{
  AddrslvBFDData *bfddata = g_new(AddrslvBFDData, 1);
  GError * tmp_error 	  = NULL;
  guint    n_installation;
  
  g_return_val_if_fail (filename, NULL);
  
  bfddata->symbol_infos = g_tree_new_full(addrslv_bfddata_address_compare,
					  NULL,
					  NULL,
					  (GDestroyNotify)addrslv_symbolinfo_unref);
  bfddata->libfile_quark  = g_quark_from_string(filename);
  bfddata->ref_count = 1;

  bfddata->virt_addr = addrslv_bfddata_run_readelf(filename, &tmp_error);
  if (tmp_error != NULL)
    {
      g_propagate_error(err, tmp_error);
      return bfddata;
    }
  
  n_installation = addrslv_bfddata_run_nm(bfddata, filename, FALSE, &tmp_error);
  if (tmp_error != NULL)
    {
      g_propagate_error(err, tmp_error);
      return bfddata;
    }
  
  if (n_installation != 0)
    return bfddata;

  n_installation = addrslv_bfddata_run_nm(bfddata, filename, TRUE, &tmp_error);
  if (n_installation == 0
      && (tmp_error != NULL))
    g_propagate_error(err, tmp_error);
  
  return bfddata;
}

static guint
addrslv_bfddata_run_nm  (AddrslvBFDData* bfddata, const gchar* filename,  gboolean dynamic, GError **err)
{
  GError * tmp_error = NULL;
  
  gchar*   nm_rawout = NULL;
  gboolean nm_spawn_return;

  gchar** nm_lines = 0;
  gchar** buffer;
  guint    n_installation;

  gchar*   nm_argv[7];  
  gchar*   nm_envp[] = {"LANG=C", NULL};

  g_return_val_if_fail (bfddata, 0);
  g_return_val_if_fail (filename, 0);

  nm_argv[0] = NM_PATH;
  nm_argv[1] = dynamic? "--dynamic": "--line-numbers";
  nm_argv[2] = "--numeric-sort";
  nm_argv[3] = "--demangle";
  nm_argv[4] = "--print-size";
  nm_argv[5] = (gchar *)filename;
  nm_argv[6] = NULL;
  
  nm_spawn_return = g_spawn_sync("/",
				 nm_argv, nm_envp,
				 /* TODO: We should return the stderr via GError. */
				 G_SPAWN_STDERR_TO_DEV_NULL, 
				 NULL, /* setup */
				 NULL, /* user data */
				 &nm_rawout, NULL,
				 NULL,
				 &tmp_error);
  if (nm_spawn_return == FALSE) {
    g_propagate_error(err, tmp_error);
    return 0;
  }

  if (nm_rawout[0] == '\0') {
    g_free(nm_rawout);
    return 0;
  }

  nm_lines 	 = g_strsplit(nm_rawout, "\n", 0);
  n_installation = 0;
  for (buffer = nm_lines; *buffer != NULL; buffer++)
      if (addrslv_bfddata_install_symbol(bfddata, *buffer, dynamic))
	n_installation++;

  g_strfreev(nm_lines);
  g_free(nm_rawout);
  nm_rawout = NULL;
  return n_installation;
}

static guint
addrslv_bfddata_run_readelf     (const gchar* filename, GError **err)
{
  GError * tmp_error = NULL;

  gchar*   readelf_rawout = NULL;
  gboolean readelf_spawn_return;  
  
  gchar** readelf_lines = 0;
  gchar** buffer;
  
  gchar*   readelf_argv[4];  
  gchar*   readelf_envp[] = {"LANG=C", NULL};

  guint virt_addr = 0;
  
  g_return_val_if_fail (filename, 0);

  readelf_argv[0] = READELF_PATH;
  readelf_argv[1] = "--program-headers";
  readelf_argv[2] = (gchar *)filename;
  readelf_argv[3] = NULL;

  readelf_spawn_return = g_spawn_sync("/",
				      readelf_argv, readelf_envp,
				      /* TODO: We should return the stderr via GError. */
				      G_SPAWN_STDERR_TO_DEV_NULL, 
				      NULL, /* setup */
				      NULL, /* user data */
				      &readelf_rawout, NULL,
				      NULL,
				      &tmp_error);
  if (readelf_spawn_return == FALSE) {
    g_propagate_error(err, tmp_error);
    return 0;
  }
  
  if (readelf_rawout[0] == '\0'  ) {
    g_free(readelf_rawout);
    return 0;
  }

  readelf_lines = g_strsplit(readelf_rawout, "\n", 0);
  for (buffer = readelf_lines; *buffer != NULL; buffer++)
    {
      gint scan_count;
      guint offset, phys_addr, filesiz, memsiz, align;

      if (*buffer[0] == '\0')
	continue;
      
      /* Type           Offset   VirtAddr   PhysAddr   FileSiz MemSiz  Flg Align */
      /* LOAD           0x000000 0x00101000 0x00101000 0x2423b8 0x2423b8 R E 0x1000 */
      /* g_message("Input: \"%s\"", *buffer); */
      scan_count = sscanf(*buffer,
			   " LOAD %x %x %x %x %x R E %x",
			  &offset, &virt_addr, &phys_addr, &filesiz, &memsiz, &align);
      /* g_message("Count: %d", scan_count); */
      if (scan_count == 6)
	break;
      virt_addr = 0;
    }
  
  /* g_message("filename: \"%s\", virt_addr: %x", filename, virt_addr); */
  
  g_strfreev(readelf_lines);
  g_free(readelf_rawout);
  readelf_rawout = NULL;
  return virt_addr;
}

#define FUNCNAME_MAX_LENGTH 128
#define FUNCNAME_ARRAY_LENGTH (FUNCNAME_MAX_LENGTH + 1)
#define SRCFILE_MAX_LENGTH PATH_MAX
#define SRCFILE_ARRAY_LENGTH (SRCFILE_MAX_LENGTH + 1)
static gboolean
addrslv_bfddata_install_symbol  (AddrslvBFDData* bfddata, 
				 const gchar* buffer,
				 gboolean dynamic)
{
  gint scan_count;
  
  guint    address = 0;
  guint    size = 0;
  gchar    symtype = 0;
  gchar    funcname[FUNCNAME_ARRAY_LENGTH];
  gchar    srcfile[SRCFILE_ARRAY_LENGTH] = {'\0'};
  gulong   lineno = 0;

  const gchar* nm_format = "%x %x %c "
    "%" G_STRINGIFY(FUNCNAME_MAX_LENGTH) "s "
    "%" G_STRINGIFY(SRCFILE_MAX_LENGTH) "[^:]:%u";
  
  const gchar* nm_dynamic_format = "%x %x %c " 
    "%" G_STRINGIFY(FUNCNAME_MAX_LENGTH) "s";
  
  AddrslvSymbolInfo * symbol_info;  

  if (dynamic)
    {
      scan_count = sscanf(buffer, 
			  nm_dynamic_format,
			  &address, &size, &symtype, funcname);
      
      if (scan_count != 4)
	return FALSE;
      
      if (symtype != 'T')
	return FALSE;
    }
  else
    {
      scan_count = sscanf(buffer,
			  nm_format,
			  &address, &size, &symtype, funcname, srcfile, &lineno);
      if ((scan_count != 4) && (scan_count != 6))
	return FALSE;

      if ((symtype != 'T')
	  && (symtype != 't'))
	return FALSE;
    }

  /* g_message ("%x: %s", address, funcname); */
  symbol_info = addrslv_symbolinfo_new(funcname, 
				       address, 
				       size,
				       (symtype == 'T')? TRUE: FALSE,
				       srcfile, 
				       lineno);

  /* g_message("[%s]%p => %s", 
	    symbol_entry->section->name,
	    GUINT_TO_POINTER(symbol_info.value), 
	    addrslv_symbol_info->funcname); */

  g_tree_insert(bfddata->symbol_infos, 
		GUINT_TO_POINTER(address),
		symbol_info);
  return TRUE;
}

static void
addrslv_bfddata_free  (AddrslvBFDData * bfddata)
{
  g_return_if_fail (bfddata);

  g_datalist_id_remove_data(&addrslv_bfds, bfddata->libfile_quark);

  g_tree_destroy(bfddata->symbol_infos);

  /* g_message("[%uF] bfd data %p", 
     bfddata->ref_count, bfddata); */

  g_free(bfddata);
}

static AddrslvBFDData*
addrslv_bfddata_ref   (AddrslvBFDData * bfddata)
{
  g_return_val_if_fail (bfddata, NULL);
  g_return_val_if_fail (bfddata->ref_count > 0, NULL);
  /* g_message("[%d++] bfd data %p", 
     bfddata->ref_count, bfddata); */
  bfddata->ref_count++;
  return bfddata;
}

static void
addrslv_bfddata_unref (AddrslvBFDData * bfddata)
{
  g_return_if_fail (bfddata);
  g_return_if_fail (bfddata->ref_count > 0);
  
  /* g_message("[%d--] bfd data %p", 
     bfddata->ref_count, bfddata); */
  bfddata->ref_count--;

  if (bfddata->ref_count == 0)
    addrslv_bfddata_free(bfddata);
}

static AddrslvSymbolInfo*
addrslv_bfddata_find  (AddrslvBFDData * bfddata, guint address)
{
  AddrslvSymbolInfo * symbol_info;
  AddrslvBfddataFindData data;
  
  g_return_val_if_fail (bfddata, NULL);
  /* g_return_val_if_fail (address, NULL); */
  
  symbol_info = g_tree_lookup (bfddata->symbol_infos, GUINT_TO_POINTER(address));
  if (symbol_info) 
    return symbol_info;
  
  data.address = address;
  data.tree    = bfddata->symbol_infos;
  return g_tree_search(bfddata->symbol_infos,
		       addrslv_bfddata_find_compare, 
		       &data);
  
}

static gint
addrslv_bfddata_find_compare (gconstpointer a, gconstpointer b)
{
  gconstpointer key = a;
  const AddrslvBfddataFindData * data = b;
  AddrslvSymbolInfo * symbol_info;
  
  if (data->address < GPOINTER_TO_UINT(key))
    return -1;

  symbol_info = g_tree_lookup (data->tree, key);
  if (symbol_info->address + symbol_info->size < data->address)
    return 1;
  else
    return 0;
}

static gint
addrslv_bfddata_address_compare(gconstpointer a,gconstpointer b, gpointer user_data)
{
  if (a < b) 
    return -1;
  else if (a > b)
    return 1;
  else
    return 0;
}

/* -------------------------------------------------------------------
   SymbolInfo
   ----------------------------------------------------------------- */ 
static AddrslvSymbolInfo*
addrslv_symbolinfo_new  (const gchar * funcname,
			 guint address,
			 guint size,
			 gboolean dynamic,
			 const gchar * filename,
			 guint lineno)
{
  /* TODO: Memleak */
  AddrslvSymbolInfo * symbol_info = g_new(AddrslvSymbolInfo, 1);

  if (funcname)
    symbol_info->funcname  = g_strdup(funcname);
  else
    symbol_info->funcname = NULL;

  symbol_info->address = address;
  symbol_info->size    = size;
  symbol_info->dynamic = dynamic;

  if (filename)
    symbol_info->srcfile_quark = g_quark_from_string(filename);
  else
    symbol_info->srcfile_quark = (GQuark)0;

  symbol_info->lineno 	 = lineno;

  symbol_info->ref_count = 1;
  
  /* g_message("[%uN] symbol info %p", 
     symbol_info->ref_count, symbol_info); */
  return symbol_info;
  
}

static void
addrslv_symbolinfo_free(AddrslvSymbolInfo * symbol_info)
{
  g_return_if_fail (symbol_info);
  
  if (symbol_info->funcname)
    g_free(symbol_info->funcname);

  /* g_message("[%uF] symbol info %p", 
     symbol_info->ref_count, symbol_info); */
  g_free(symbol_info);
}

AddrslvSymbolInfo * 
addrslv_symbolinfo_ref   (AddrslvSymbolInfo * symbol_info)
{
  g_return_val_if_fail (symbol_info, NULL);
  g_return_val_if_fail (symbol_info->ref_count > 0, NULL);

  /* g_message("[%u++] symbol info %p", 
     symbol_info->ref_count, symbol_info); */
  symbol_info->ref_count++;
  return symbol_info;
}

void
addrslv_symbolinfo_unref (AddrslvSymbolInfo * symbol_info)
{
  g_return_if_fail (symbol_info);
  g_return_if_fail (symbol_info->ref_count > 0);

  /* g_message("[%u--] symbol info %p", 
     symbol_info->ref_count, symbol_info); */
  symbol_info->ref_count--;

  if (symbol_info->ref_count == 0)
    addrslv_symbolinfo_free(symbol_info);
}

static gboolean
addrslv_symbolinfo_search_by_funcname (gpointer key,
				       gpointer value,
				       gpointer data)
{
  const gchar * funcname 	  = *(const gchar**)data;
  AddrslvSymbolInfo * symbol_info = value;
  
  g_return_val_if_fail (key, TRUE);
  g_return_val_if_fail (symbol_info, TRUE);
  g_return_val_if_fail (funcname, TRUE);

  if (strcmp(symbol_info->funcname, funcname) == 0)
    {
      *((AddrslvSymbolInfo **)data) = symbol_info;
      return TRUE;
    }
  else
    return FALSE;
}

/* -------------------------------------------------------------------
   PAttach: Data Attached to PMaps 
   ----------------------------------------------------------------- */ 
static AddrslvPMapsAttachedData*
addrslv_pattach_get     (PMaps * pmaps)
{
  AddrslvPMapsAttachedData* pattach;

  pattach = addrslv_pattach_try_get(pmaps);
  if (pattach == NULL)
    {
      pattach = addrslv_pattach_new(pmaps);
      pmaps_set_data(pmaps, PATTACH_KEY, pattach, 
		     (GDestroyNotify)addrslv_pattach_free);
    }
  return pattach;
}

static AddrslvPMapsAttachedData*
addrslv_pattach_try_get (PMaps * pmaps)
{
  return pmaps_get_data(pmaps, PATTACH_KEY);
}

static AddrslvPMapsAttachedData*
addrslv_pattach_new(PMaps * pmaps)
{
  AddrslvPMapsAttachedData * pattach = g_new(AddrslvPMapsAttachedData, 1);

  pattach->pmaps            = pmaps;
  pattach->hooked_to_update = FALSE;
  pattach->hook_id          = 0;
  pattach->gc_marked_bfddata = NULL;
  return pattach;
}

static void
addrslv_pattach_free(AddrslvPMapsAttachedData* pattach)
{
  g_return_if_fail (pattach->gc_marked_bfddata == NULL);
  if (pattach->hooked_to_update)
    pmaps_remove_update_notify(pattach->pmaps, pattach->hook_id);
  g_free(pattach);
}

static void
addrslv_pattach_gc_mark (AddrslvPMapsAttachedData* pattach, 
			 AddrslvBFDData* bfddata)
{
  g_return_if_fail (pattach);
  g_return_if_fail (bfddata);

  addrslv_bfddata_ref(bfddata);
  pattach->gc_marked_bfddata = g_slist_prepend(pattach->gc_marked_bfddata, 
					       bfddata);
  
}

static void
addrslv_pattach_gc_sweep (AddrslvPMapsAttachedData* pattach)
{
  GSList * list = pattach->gc_marked_bfddata;
  while (list)
    {
      addrslv_bfddata_unref(list->data);
      list = g_slist_next(list);
    }
  g_slist_free(pattach->gc_marked_bfddata);
  pattach->gc_marked_bfddata = NULL;
}

/* 
 * GError related
 */
GQuark
addrslv_bfd_error_quark (void)
{
  static GQuark quark = 0;

  if (!quark)
    quark = g_quark_from_static_string ("addrslv_bfd_error");
  
  return quark;
}

/* arch-tag: dc19df42-ccfc-4825-aa09-e69eedef1390
   (do not change this comment) */
