/* backtrace.c --- Getting backtrace for x86

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include <backtrace.h>
#include <pmaps.h>

#include <sys/types.h>
#include <unistd.h>

#define return_address(ebp)  *(unsigned long *)(((char *)(ebp)) + 4)
#define lower_stackbase(ebp) *(((unsigned long *)(ebp)))

static gboolean backtrace_ebp_verify(guint ebp);

GSList *
backtrace_funcall(void)
{
  GSList * bt = NULL;
  guint ebp = 0;
  guint raddress;
  
  asm volatile("movl %%ebp,%0": "=m"(ebp): : "memory");

  if (ebp == 0)
    goto out;
  raddress = return_address(ebp);
  /* Just throw away this return address.
     This is the bottom return address: the address
     of function which calls this `backtrace_funcall'. */
  ebp = lower_stackbase(ebp);
  
  while (ebp != 0)
    {
      if (backtrace_ebp_verify(ebp) == FALSE) 
	goto out;
      raddress = return_address(ebp);
      bt = g_slist_prepend(bt, GUINT_TO_POINTER(raddress));
      ebp = lower_stackbase(ebp);
    }

 out:
  return bt;
}

/* TRUE if ebp points somewhere in stacks. */
static gboolean backtrace_ebp_verify_internal(PMaps * pmaps,
					      guint ebp);
static gboolean
backtrace_ebp_verify(guint ebp)
{
  static PMaps *  pmaps = NULL;
  static gboolean pmaps_error = FALSE;

  gboolean verify_ok;
  gboolean updating_ok;
  
  if (pmaps_error == TRUE) 
    return FALSE;
  
  if (pmaps == NULL) 
    {
      pid_t pid;
      
      pid = getpid();
      pmaps = pmaps_new(pid, NULL);
      
    }

  if (pmaps == NULL)
    {
      pmaps_error = TRUE;
      return FALSE;
    }
  
  verify_ok = backtrace_ebp_verify_internal(pmaps, ebp);
  
  if (verify_ok == TRUE)
    {
      /* Maybe ok....Strictly speaking, updating
       is needed. However, it takes longer time.*/
      return TRUE;
    }

  updating_ok = pmaps_update(pmaps, NULL);
  if (updating_ok == FALSE)
    {
      pmaps_error = TRUE;
      return FALSE;
    }
  verify_ok = backtrace_ebp_verify_internal(pmaps, ebp);
  if (verify_ok == FALSE) 
    g_message ("ebp: %p points wrong place", GUINT_TO_POINTER(ebp));
  
  return verify_ok;
}

static gboolean
backtrace_ebp_verify_internal(PMaps * pmaps,
			      guint ebp)
{
  static PMapsEntry * entry = NULL;

  /* TODO
     pmaps entry could be replaced with a stack base. */
  if (entry)
    {
      if (pmaps_entry_include(entry, ebp) == 0)
	return TRUE;
      pmaps_entry_unref(entry);
    }
	
  entry = pmaps_find(pmaps, ebp);

  if (!entry) 
    return FALSE;

  if (entry->filename[0] != '\0')
    {
      entry = NULL;
      return FALSE;
    }

  if (!(entry->prot & PMAPS_READABLE))
    {
      entry = NULL;
      return FALSE;
    }

  if (!(entry->prot & PMAPS_WRITABLE))
    {
      entry = NULL;
      return FALSE;
    }

#if 0
  /*
    $ cat /proc/22083/maps
    ...
    bf596000-bf59c000 rw-p 00000000 00:00 0
    bf5b6000-bf5b7000 rw-p 00000000 00:00 0
 => bffb8000-bffc8000 rwxp fffe8000 00:00 0
    bffc8000-c0000000 rw-p ffff8000 00:00 0
    ...very strange
  */
  if (entry->prot & PMAPS_EXECUTABLE)
    {
      entry = NULL;
      return FALSE;
    }
#endif  /* 0 */

  if (entry->prot & PMAPS_SHARED)
    {
      entry = NULL;
      return FALSE;
    }

  entry = pmaps_entry_ref(entry);
  return TRUE;
}
/* arch-tag: 12d823be-06f3-42e3-a10e-c314be7830c4
   (do not change this comment) */
