/* fptreel.c --- Reeling up a function address from a pointer to it

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "fptreel.h"
#include <addrslv.h>
#include <stdlib.h>

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif  /* G_LOG_DOMAIN */
#define FPTREEL_LOG_DOMAIN "LibFptReel"
#define G_LOG_DOMAIN FPTREEL_LOG_DOMAIN

guint
fptreel_reelup (gpointer bait, PMaps * pmaps, GError ** err)
{
  return fptreel_reelup_debug(bait, pmaps, NULL, err);
}

#define fptreel_return(return_value, stage_value) G_STMT_START { 	\
    if (stage)								\
      *stage = (stage_value);						\
    return (return_value);						\
  } G_STMT_END
 
/* 
 * This is for NetBSD
 */
#define NETBSD__LIBEXEC__LD_ELF_SO "/usr/libexec/ld.elf_so"
#define RTLD_BIND     "_rtld_bind"
#define RTLD_BIND_KEY "ftreel_rtld_bind"
/* TODO */

/* 
 * This is for Linux/Libc2/ld.2.so.
 */
#define LINUX__LIBC_2__LD_2_SO "/lib/ld-2"
#define FIXUP "fixup"
#define FIXUP_KEY "fptreel_fixup" 

static gpointer fptreel_find_fixup (PMaps * pmaps, GError ** err);
static void fptreel_find_ld_so_pmaps_entry(PMapsEntry * entry, gpointer callback_data);

static gpointer
fptreel_find_fixup (PMaps * pmaps, GError ** err)
{
  gpointer fixup;
  PMapsEntry * fixup_pentry = NULL;
  AddrslvSymbolInfo* fixup_symbol_info;
  GError * tmp_error = NULL;
  
  g_return_val_if_fail (pmaps, NULL);

  fixup = pmaps_get_data(pmaps, FIXUP_KEY);
  if (fixup)
    return fixup;
  
  pmaps_foreach(pmaps,
		fptreel_find_ld_so_pmaps_entry,
		&fixup_pentry);
  if (fixup_pentry == NULL)
    return NULL;

  fixup_symbol_info = addrslv_reverse_resolve(pmaps,
					      fixup_pentry,
					      FIXUP,
					      &tmp_error);
  if (fixup_symbol_info == NULL)
    {
      g_propagate_error (err, tmp_error);
      return NULL;
    }

  pmaps_set_data(pmaps, FIXUP_KEY, 
		 GUINT_TO_POINTER(fixup_symbol_info->address),
		 NULL);
  
  return GUINT_TO_POINTER(fixup_symbol_info->address);
}


static void
fptreel_find_ld_so_pmaps_entry(PMapsEntry * entry, gpointer callback_data)
{
  if (entry->filename
      && (g_strrstr(entry->filename,
		    LINUX__LIBC_2__LD_2_SO))
      && (entry->prot & PMAPS_EXECUTABLE))
    *((PMapsEntry **)callback_data) = entry;
}


guint
fptreel_reelup_debug (gpointer bait, PMaps * pmaps, guint * stage, GError ** err)
{
  guint   fish;
  guchar* reel;

  guchar jmp_near_abs_indir_opcode = 0xFF;
  /* 0x25
     7 6     5 4 3    2 1 0
     mod  Reg/Opcode   R/M
     0 0     1 0 0    1 0 1
      |       => /4     |
      |                 |
      +-----------------+ => disp32 */
  guchar jmp_disp32_effective_addr = 0x25;
  
  guchar push_imm32_opcode = 0x68;

  guchar push_disp32_opcode = 0xFF;
  /* 0x35
     7 6     5 4 3    2 1 0
     mod  Reg/Opcode   R/M
     0 0     1 1 0    1 0 1
      |       => /6     |
      |                 |
      +-----------------+ => disp32 */
  guchar push_disp32_effective_addr = 0x35;

  guchar jmp_near_rel_disp_opcode = 0xE9;

  guchar * plt;
  GError * tmp_error = NULL;
  
  guint edx;			/* eax? */
  guint eax;			/* edx? */

  reel = (guchar *)bait;
  
  if (reel[0] != jmp_near_abs_indir_opcode)
    fptreel_return(GPOINTER_TO_UINT(bait), 1);

  
  if (reel[1] != jmp_disp32_effective_addr)
    fptreel_return(GPOINTER_TO_UINT(bait), 2);
  
  fish = ((  (guint)reel[5]) << 24
	  | ((guint)reel[4]) << 16
	  | ((guint)reel[3]) << 8
	  | ((guint)reel[2]));

  if (g_getenv("LD_BIND_NOW") != NULL)
    fptreel_return(*(guint *)GUINT_TO_POINTER(fish), 3);

  reel = (guchar *)*(guint *)fish;
  if (reel[0] != push_imm32_opcode)
    fptreel_return(*(guint *)GUINT_TO_POINTER(fish), 4);

  edx = (( ((guint)reel[4])  << 24)
	 | (((guint)reel[3]) << 16)
	 | (((guint)reel[2]) << 8)
	 | (((guint)reel[1]) << 0));

  reel = &(reel[5]);
  if (reel[0] != jmp_near_rel_disp_opcode)
    fptreel_return(GPOINTER_TO_UINT(bait), 5);

  plt  = (guchar *)(GPOINTER_TO_UINT(&reel[5])
		    + ((  (gint)reel[4]) << 24
		       | ((gint)reel[3]) << 16
		       | ((gint)reel[2]) << 8
		       | ((gint)reel[1])));

  reel = plt;
  if (reel[0] != push_disp32_opcode)
    fptreel_return(GPOINTER_TO_UINT(bait), 6);

  if (reel[1] != push_disp32_effective_addr)
    fptreel_return(GPOINTER_TO_UINT(bait), 7);
  
  eax = *(guint *)(( ((guint)reel[5])  << 24)
		   | (((guint)reel[4]) << 16)
		   | (((guint)reel[3]) << 8)
		   | (((guint)reel[2]) << 0));
  {
    static guint (*fixup) (guint a, guint d) __asm__("fptreel_fixup")  = NULL;

    fixup = fptreel_find_fixup(pmaps, &tmp_error);

    if (tmp_error) 
      {
	g_propagate_error (err, tmp_error);
	fptreel_return(GPOINTER_TO_UINT(bait), 8);
      }
    if (fixup == 0)
      fptreel_return(GPOINTER_TO_UINT(bait), 9);

    /* TODO: Don't use a symbol as call's operand. Passing it 
       via a register.*/
    __asm__ volatile ("call *fptreel_fixup"
		      : "=a" (fish) : "0" (eax), "d" (edx));
    
    fptreel_return(GPOINTER_TO_UINT(fish), 0);
  }
}

/* arch-tag: 62343275-5c34-4747-842f-605e8ec2ed71
   (do not change this comment) */
