/* mainpos.c --- Finding the pmaps entry which may include `main'

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "mainpos.h"
#include <string.h>

static void     mainpos_explorer(PMapsEntry * entry, gpointer user_data);

typedef gboolean (* MainposDetector)(const gchar * filename, GError ** err);
typedef struct _MainposKnapsack MainposKnapsack;
struct _MainposKnapsack
{
  MainposDetector detector;
  PMapsEntry * main_entry;
};

static gboolean mainpos_find___libc_start_main(const gchar * filename, GError ** err);

static gboolean mainpos_find_line(gchar **argv, gchar * needle, gboolean partial, GError ** err);
static gboolean mainpos_find_line_by_nm(const gchar * filename, gchar * needle, gboolean partial, GError ** err);


PMapsEntry *
mainpos_find(PMaps * pmaps, GError ** err)
{
  MainposKnapsack knapsack;
  
  g_return_val_if_fail (pmaps, NULL);
  
  knapsack.main_entry = NULL;
  knapsack.detector = mainpos_find___libc_start_main;
  pmaps_foreach(pmaps, mainpos_explorer, &knapsack);
  return knapsack.main_entry;
}

static void
mainpos_explorer(PMapsEntry * entry, gpointer user_data)
{
  MainposKnapsack * knapsack = user_data;
  
  g_return_if_fail (entry);
  g_return_if_fail (entry->filename);
  g_return_if_fail (user_data);

  /* TODO: Warn double hits. */
  if (knapsack->main_entry) 
    return ;
  
  if ((entry->prot & PMAPS_EXECUTABLE)
      && knapsack->detector(entry->filename, NULL))
    knapsack->main_entry = entry;
}

static gboolean
mainpos_find___libc_start_main(const gchar * filename, GError ** err)
{
  return mainpos_find_line_by_nm(filename, "U __libc_start_main", TRUE, err);
}

static gboolean
mainpos_find_line_by_nm(const gchar * filename, gchar * needle, gboolean partial, GError ** err)
{
  gchar* argv[4];
  argv[0] = NM_PATH;
  argv[1] = "--dynamic";
  argv[2] = (gchar *)filename;
  argv[3] = NULL;
  return mainpos_find_line(argv, needle, partial, err);
}

/** mainpos_find_line
    Use strstr if partial is non-zero.
    Use strcmp if partial is zero. */
static gboolean
mainpos_find_line(gchar **argv, gchar * needle, gboolean partial, GError ** err)
{
  gchar*   envp[] = {"LANG=C", NULL};
  gboolean cmd_return;
  gchar*   cmd_rawout = NULL;
  gchar**  cmd_lines  = NULL;
  GError * tmp_error  = NULL;
  
  gchar** line;
  
  g_return_val_if_fail (argv, FALSE);
  g_return_val_if_fail (needle, FALSE);

  cmd_return = g_spawn_sync("/",
			    argv, envp,
			    /* TODO: We should return the stderr via GError. */
			    G_SPAWN_STDERR_TO_DEV_NULL, 
			    NULL,
			    NULL,
			    &cmd_rawout, NULL,
			    NULL,
			    &tmp_error);
  if (cmd_return == FALSE) 
    {
      g_propagate_error(err, tmp_error);
      return FALSE;
    }

  if (cmd_rawout[0] == '\0') {
    g_free(cmd_rawout);
    return FALSE;
  }

  cmd_lines = g_strsplit(cmd_rawout, "\n", 0);
  for (line = cmd_lines; *line != NULL; line++)
    {
      if (partial) {
	if (strstr(*line, needle) != NULL) 
	  return TRUE;
      } else {
	if (strcmp(*line, needle) == 0) 
	  return TRUE;
      }
    }
  return FALSE;
}

/* arch-tag: 07a565b2-60fa-4e9c-ba7f-f7409ea387c8
   (do not change this comment) */
