/* gset.c --- unordered pointer container

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "gset.h"

struct _GSet
{
  GSetCopyFunc copy_func;
  GTree*       tree;
};

typedef struct _GSetForeachData GSetForeachData;
struct _GSetForeachData
{
  GFunc func;
  gpointer data;
};

static gint g_set_compare_data(gconstpointer a, gconstpointer b, gpointer user_data);
static gint g_set_compare_wildcard(gconstpointer a, gconstpointer b) {return 0;}
static gboolean g_set_traverse(gpointer key, gpointer value, gpointer data);

GSet*
g_set_new      (GCompareFunc    compare_func,
		GSetCopyFunc    copy_func,
		GDestroyNotify  destroy_func)
{
  GSet * set 	 = g_new(GSet, 1);
  set->copy_func = copy_func;
  set->tree 	 = g_tree_new_full(g_set_compare_data,
				   compare_func, 
				   destroy_func,
				   NULL);
  return set;
}

static gint
g_set_compare_data(gconstpointer a, gconstpointer b, gpointer user_data)
{
  GCompareFunc compare_func = user_data;
  return compare_func(a, b);
}

void
g_set_destroy  (GSet * set)
{
  g_return_if_fail (set);
  g_return_if_fail (set->tree);
  g_tree_destroy(set->tree);
  g_free(set);
}

void
g_set_add      (GSet * set, gconstpointer data)
{
  gpointer copied;
  g_return_if_fail (set);
  if (set->copy_func) 
    copied = set->copy_func(data);
  else
    copied = (gpointer)data;
  g_tree_insert(set->tree, copied, copied);
}

gboolean
g_set_contains (GSet * set, gconstpointer data)
{
  g_return_val_if_fail (set, FALSE);
  return g_tree_lookup(set->tree, data)? TRUE: FALSE;
}

void
g_set_remove     (GSet * set, gconstpointer data)
{
  g_return_if_fail (set);
  g_tree_remove(set->tree, data);
}

gpointer
g_set_choose_any (GSet * set)
{
  g_return_val_if_fail (set, NULL);
  return g_tree_search(set->tree, g_set_compare_wildcard, NULL);
}

gpointer
g_set_pick_any   (GSet * set)
{
  gpointer data, copied;
  g_return_val_if_fail (set, NULL);
 
  data = g_tree_search(set->tree, g_set_compare_wildcard, NULL);
  if (data == NULL) 
    return NULL;

  if (set->copy_func) 
      copied = set->copy_func(data);
  else
    copied = data;

  g_set_remove(set, data);
  return copied;
}

guint
g_set_count    (GSet * set)
{
  g_return_val_if_fail (set, 0);
  return g_tree_nnodes(set->tree);
}

void
g_set_foreach  (GSet            *set,
		GFunc            func,
		gpointer	 user_data)
{
  GSetForeachData user_request;  

  g_return_if_fail (set);
  g_return_if_fail (func);

  user_request.func = func;
  user_request.data = user_data;
  g_tree_foreach(set->tree, g_set_traverse, &user_request);
}

static gboolean
g_set_traverse(gpointer key, gpointer value, gpointer data)
{
  GSetForeachData * user_request = data;
  user_request->func(key, user_request->data);
  return FALSE;
}

/* arch-tag: 0aebca38-85b1-4f8f-b1d6-c7621f0b4ecb
   (do not change this comment) */
