/* backtrace_viewer.c --- backtrace viewer implementation

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "service.h"
#include "utils.h"

#include "i18n.h"

#include <addrslv.h>

#include <ginspector/ginspector-gtk-tag.h>
#include <ginspector/ginspector-tag.h>

#include <gtk/gtk.h>

typedef struct _GTracerBacktraceView GTracerBacktraceView;

#define CONTEXT_ID_KEY "gtracer_backtrace_context_id"
enum
{
  COL_ADDRESS = 0,
  COL_SYMBOL  = 1,
  NUM_COLS
};

static GtkTreeModel * gtrace_backtrace_model_new(GSList * bt);
static GtkWidget *    gtrace_backtrace_viewer_new(GtkTreeModel * model);
static GtkWidget *    gtrace_backtrace_viewer_view_new(void);

static void           gtracer_backtrace_viewer_selection_changed(GtkTreeSelection *selection,
								 gpointer user_data);

static void           gtracer_backtrace_model_extract_address (GtkTreeViewColumn *col,
							       GtkCellRenderer   *renderer,
							       GtkTreeModel      *model,
							       GtkTreeIter       *iter,
							       gpointer           user_data);
static void           gtracer_backtrace_model_extract_function (GtkTreeViewColumn *col,
								GtkCellRenderer   *renderer,
								GtkTreeModel      *model,
								GtkTreeIter       *iter,
								gpointer           user_data);

static gchar*         gtracer_backtrace_model_extract_message    (GtkTreeModel * model,
								  GtkTreeIter *iter);

void
gtrace_backtrace_show_backtrace(const gchar * header, GSList * bt)
{
  GtkWidget * window;
  GtkWidget * viewer;
  gchar * title;
  
  GtkTreeModel * model;

  model  = gtrace_backtrace_model_new(bt);
  viewer = gtrace_backtrace_viewer_new(model);
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  g_inspector_tag_check_widget(window);
  gtk_container_add(GTK_CONTAINER(window), viewer);

  if (header)
    title = g_strdup_printf(_("Backtrace: %s"), header);
  else
    title = g_strdup(_("Backtrace"));

  gtk_window_set_title(GTK_WINDOW(window), title);
  g_free(title);
  
  gtk_widget_show_all(window);
}

static GtkTreeModel *
gtrace_backtrace_model_new(GSList * bt)
{
  GSList *frame;
  GtkListStore * store;
  GtkTreeIter  iter;
  
  store = gtk_list_store_new(NUM_COLS, G_TYPE_UINT, AddrslvSymbolInfoType);
  for(frame = bt; frame; frame = frame->next) 
    {
      AddrslvSymbolInfo * symbol_info;
      
      symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(frame->data), 
						  NULL);
      gtk_list_store_append (store, &iter);
      gtk_list_store_set (store, &iter, 
			  COL_ADDRESS, GPOINTER_TO_UINT(frame->data),
			  COL_SYMBOL, symbol_info, 
			  -1);
    }
  return GTK_TREE_MODEL(store);
}

static GtkWidget *
gtrace_backtrace_viewer_new(GtkTreeModel * model)
{
  GtkWidget * vbox;
  GtkWidget * scrolled_window;
  GtkAdjustment * adjustment;
  GtkWidget * tree_view;
  GtkTreeSelection *tree_selection;
  GtkWidget * statusbar;
  guint context_id;
  
  vbox = gtk_vbox_new(FALSE, 4);
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_NEVER, 
				 GTK_POLICY_AUTOMATIC);
  adjustment 	  = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));
  adjustment 	  = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));

  gtk_box_pack_start(GTK_BOX(vbox), 
		     scrolled_window, 
		     TRUE, 	/* expand */
		     TRUE,	/* fill */
		     2);
  
  tree_view = gtrace_backtrace_viewer_view_new();
  gtk_widget_set_size_request (tree_view, -1, 200);
  gtk_tree_view_set_model (GTK_TREE_VIEW (tree_view), model);
  gtk_container_add(GTK_CONTAINER(scrolled_window), tree_view);

  statusbar  = gtk_statusbar_new();
  context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
					    "Backtrace extra information");
  g_object_set_data(G_OBJECT(statusbar), CONTEXT_ID_KEY, GUINT_TO_POINTER(context_id));
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, "");
  gtk_box_pack_start(GTK_BOX(vbox), statusbar, 
		     FALSE, 	/* expand */
		     FALSE,	/* fill */
		     2);
  gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), 
				    FALSE);

  tree_selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
  g_signal_connect (tree_selection, 
		    "changed", 
		    G_CALLBACK (gtracer_backtrace_viewer_selection_changed), 
		    statusbar);
  
  return vbox;
}

static GtkWidget *
gtrace_backtrace_viewer_view_new(void)
{
  GtkWidget * tree_view;
  GtkCellRenderer * text_renderer;
  gint position;
  GtkTreeViewColumn * col;
  GtkTreeSelection *tree_selection;

  tree_view = gtk_tree_view_new();
  text_renderer = gtk_cell_renderer_text_new ();
  g_inspector_tag_to_object(G_OBJECT(text_renderer));
  
  position = 0;
  position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree_view),
							position,
							_("Address"),
							text_renderer,
							gtracer_backtrace_model_extract_address,
							NULL,
							NULL);
  col = gtk_tree_view_get_column(GTK_TREE_VIEW(tree_view), position - 1);
  gtracer_register_tree_view_column(col);

  position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree_view),
							position,
							_("Function"),
							text_renderer,
							gtracer_backtrace_model_extract_function,
							NULL,
							NULL);
  col = gtk_tree_view_get_column(GTK_TREE_VIEW(tree_view), position - 1);
  gtracer_register_tree_view_column(col);

  tree_selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
  g_inspector_tag_to_object(G_OBJECT(tree_selection));
  
  gtk_tree_selection_set_mode (GTK_TREE_SELECTION (tree_selection),
			       GTK_SELECTION_SINGLE);

  return tree_view;
}

static void
gtracer_backtrace_viewer_selection_changed(GtkTreeSelection *selection,
					   gpointer user_data)
{
  GtkWidget * statusbar;
  GtkTreeModel * model;
  GtkTreeIter iter;
  gchar * msg;
  gpointer context_id;
  
  g_return_if_fail (selection);
  g_return_if_fail (user_data);

  if (gtk_tree_selection_get_selected(selection, &model, &iter) == TRUE)
    {
      statusbar  = user_data;
      context_id = g_object_get_data(G_OBJECT(statusbar), CONTEXT_ID_KEY);
      gtk_statusbar_pop(GTK_STATUSBAR(statusbar), GPOINTER_TO_UINT(context_id));
      msg = gtracer_backtrace_model_extract_message(model, &iter);
      if (msg) 
	{
	  gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
			     GPOINTER_TO_UINT(context_id),
			     msg);
	  g_free(msg);
	}
      else 
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
			   GPOINTER_TO_UINT(context_id),
			   "");
      
    }
}

static void
gtracer_backtrace_model_extract_address (GtkTreeViewColumn *col,
					 GtkCellRenderer   *renderer,
					 GtkTreeModel      *model,
					 GtkTreeIter       *iter,
					 gpointer           user_data)
{
  guint address;
#define BUFFER_LENGTH 32
  gchar buffer[BUFFER_LENGTH];
  
  gtk_tree_model_get(model, iter, COL_ADDRESS, &address, -1);
  snprintf(buffer, BUFFER_LENGTH, "%08x", address);  
  g_object_set(renderer, "text", buffer, NULL);
}


static void
gtracer_backtrace_model_extract_function (GtkTreeViewColumn *col,
					  GtkCellRenderer   *renderer,
					  GtkTreeModel      *model,
					  GtkTreeIter       *iter,
					  gpointer           user_data)
{
  AddrslvSymbolInfo * symbol_info;
  
  gtk_tree_model_get(model, iter, COL_SYMBOL, &symbol_info, -1);
  if (symbol_info && symbol_info->funcname)
    {
      g_object_set(renderer, "text", symbol_info->funcname, NULL);
      g_boxed_free(AddrslvSymbolInfoType, symbol_info);
    }
  else
    g_object_set(renderer, "text", _("{static, no debug information}"), NULL);
}

static gchar*
gtracer_backtrace_model_extract_message    (GtkTreeModel * model,
					    GtkTreeIter *iter)
{
  guint address;
  AddrslvSymbolInfo * symbol_info;
  
  gchar * msg = NULL;;

  gtk_tree_model_get(model, iter, 
		     COL_ADDRESS, &address,
		     COL_SYMBOL, &symbol_info, -1);

  msg = gtracer_pmaps_message_new(symbol_info, address);
  if (symbol_info)
    g_boxed_free(AddrslvSymbolInfoType, symbol_info);	/* MODEL_GET_FREE */
  return msg;
}

/* arch-tag: e9500ce5-bb52-44d1-bd9c-327286b88ab0
   (do not change this comment) */
