/* bubble.c --- Move a list element like bubble(surfacing and sinking)

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "bubble.h"

GList *
g_list_surface (GList * head, GList * sub, GCompareFunc func)
{
  GList * larger;
  
  g_return_val_if_fail (head, NULL);
  g_return_val_if_fail (sub, head);
  g_return_val_if_fail (func, head);

  if (g_list_previous(sub) == NULL) 
    /* sub is the head of the list.
       No place to move more. */
    return head;
  
  /* Find one larger than sub */
  for(larger = g_list_previous(sub); 
      larger && (func(sub->data, larger->data) < 0);
      larger = g_list_previous(larger))
    ;


  if (larger == g_list_previous(sub))
    /* The larger one is at the next of sub. 
       => Nothing to do. */
    ;			
  else if (larger == NULL) 
    {
      /* sub is the larget one in the list starting
	 from head. Move the sub to head. */
      // (void)g_list_remove_link(g_list_previous(sub), sub); 
      head = g_list_remove_link(head, sub); 
      head = g_list_concat(sub, head);
    } 
  else
    {
      /* Insert sub after the `larger'. */
      GList * tmp;
      
      // (void)g_list_remove_link(g_list_previous(sub), sub);

      head = g_list_remove_link(head, sub);
      tmp  = g_list_next(larger);
      
      larger->next  = NULL;
      tmp->prev = NULL;

      g_list_concat(larger, sub);
      g_list_concat(sub, tmp);
    }
  return head;
}

GList *
g_list_sink (GList * head, GList * sub, GCompareFunc func)
{
  GList * smaller;
  
  g_return_val_if_fail (head, NULL);
  g_return_val_if_fail (sub, head);
  g_return_val_if_fail (func, head);

  if (g_list_next(sub) == NULL) 
    /* sub is the tail of the list.
       No place to move more. */
    return head;
  
  /* Find one smaller than sub */
  for(smaller = g_list_next(sub); 
      smaller && (func(sub->data, smaller->data) > 0);
      smaller = g_list_next(smaller))
    ;


  if (smaller == g_list_next(sub))
    /* The smaller one is at the next of sub. 
       => Nothing to do. */
    ;			
  else if (smaller == NULL) 
    {
      /* sub is the smallest one in the list starting
	 from head. */
      head = g_list_remove_link(head, sub);
      head = g_list_concat(head, sub);
    } 
  else
    {
      /* Insert sub before the `smaller'. */
      GList * tmp;

      head = g_list_remove_link(head, sub);

      tmp = g_list_previous(smaller);
      
      smaller->prev  = NULL;
      tmp->next = NULL;

      g_list_concat(tmp, sub);
      g_list_concat(sub, smaller);
    }
  return head;
}

/*
 * For testing
 */
/* gcc -DBUBBLE_TEST -O0 -g bubble.c \
       `pkg-config --libs --cflags glib-2.0` -fprofile-arcs -ftest-coverage */
#ifdef BUBBLE_TEST 
typedef struct _Bubble
{
  char * name;
  guint diff;
  GList * list;
} Bubble;

Bubble *
bubble_new(char * name, guint diff)
{
  Bubble * bubble;

  bubble       = g_new(Bubble, 1);
  bubble->name = g_strdup(name);
  bubble->diff = diff;
  bubble->list = NULL;
  return bubble;
}

gint
bubble_comp(gconstpointer a,
	    gconstpointer b)
{
  const Bubble * bubble_a = a;
  const Bubble * bubble_b = b;
  
  if (bubble_a->diff < bubble_b->diff)
    return -1;
  else if (bubble_a->diff == bubble_b->diff)
    return 0;
  else
    return 1;
}

void
dump (GList * list)
{
  while (list) 
    {
      Bubble * data;
      data = list->data;
      g_message ("%s: %d",
		 data->name,
		 data->diff);
      list = list->next;
    }
}

int
main(int argc, char ** argv)
{
  GList * list 	      	= NULL;
  Bubble * bubble_gcc 	= bubble_new("gcc", 3);
  Bubble * bubble_gdb 	= bubble_new("gdb", 3);
  Bubble * bubble_emacs = bubble_new("emacs", 3);
  Bubble * bubble_tex = bubble_new("tex", 3);
  
  list 		   = g_list_insert_sorted(list, bubble_gcc, bubble_comp);
  bubble_gcc->list = g_list_find(list, bubble_gcc);
  
  list = g_list_insert_sorted(list, bubble_gdb, bubble_comp);
  bubble_gdb->list = g_list_find(list, bubble_gdb);

  list = g_list_insert_sorted(list, bubble_emacs, bubble_comp);
  bubble_emacs->list = g_list_find(list, bubble_emacs);

  list = g_list_insert_sorted(list, bubble_tex, bubble_comp);
  bubble_tex->list = g_list_find(list, bubble_tex);

  bubble_gcc->diff++;
  list = g_list_surface(list, bubble_gcc->list, bubble_comp);

  bubble_gcc->diff++;
  list = g_list_surface(list, bubble_gcc->list, bubble_comp);

  bubble_tex->diff += 3;
  list = g_list_surface(list, bubble_tex->list, bubble_comp);

  bubble_gcc->diff -= 3;
  list = g_list_sink(list, bubble_gcc->list, bubble_comp);

  bubble_gcc->diff -= 1;
  list = g_list_sink(list, bubble_gcc->list, bubble_comp);

  bubble_tex->diff -= 1;
  list = g_list_sink(list, bubble_tex->list, bubble_comp);

  bubble_gcc->diff += 3;
  list = g_list_surface(list, bubble_gcc->list, bubble_comp);

  bubble_tex->diff = 1;
  list = g_list_sink(list, bubble_tex->list, bubble_comp);

  bubble_emacs->diff = 2;
  list = g_list_sink(list, bubble_emacs->list, bubble_comp);

  bubble_gcc->diff = 2;
  list = g_list_sink(list, bubble_gcc->list, bubble_comp);
  
  dump(list);  return 0;
} 


#endif /* Def: BUBBLE_TEST */

/* arch-tag: 44fc5193-f98c-47c2-99cd-7e951bd02878
   (do not change this comment) */
