/* environ_viewer.c --- environment variable viewer

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "environ_service.h"
#include "plugin.h"
#include "utils.h"

#include "i18n.h"

#include <gtk/gtk.h>
#include <ginspector/ginspector-tag.h>
#include <ginspector/ginspector-gtk-tag.h>
#include "ginspector/ginspector-procedure.h"

typedef struct _GTracerEnvironViewer GTracerEnvironViewer;
struct _GTracerEnvironViewer
{
  GtkWidget * window;
  GtkWidget * view;
  GtkWidget * statusbar;
  GtkTreeSelection   * selection;
};

static GTracerEnvironViewer * gtracer_environ_viewer_new    (void);
static gpointer               gtracer_environ_viewer_show   (gpointer data,
							     gpointer func_data, 
							     gpointer invoke_data);
static void                   gtracer_environ_viewer_update (GTracerEnvironViewer * viewer);
static void                   gtracer_environ_viewer_unset  (GTracerEnvironViewer * viewer);
static void                   gtracer_environ_viewer_add    (GTracerEnvironViewer * viewer);

static void                   gtracer_environ_viewer_value_edited(GtkCellRendererText * cell,
								  gchar               * path_string,
								  gchar               * new_text,
								  gpointer              user_data);

enum
{
  SORTID_PAIR = 0,
} ;

void
gtracer_environ_viewer_init(GError ** err)
{
  GTracerEnvironViewer * viewer;
  viewer = gtracer_environ_viewer_new();

  g_inspector_register_procedure ("environ-viewer",
				  "Environ Viewer",
				  G_INSPECTOR_PROCEDURE_TYPE_WINDOW,
				  G_INSPECTOR_ELEMENT_TYPE_NONE,
				  G_TYPE_NONE,
				  gtracer_environ_viewer_show,
				  viewer);
  g_tracer_append_menu_item(_("/View/Environment Variables"), 
			    NULL, 
			    "environ-viewer");
}

static GTracerEnvironViewer *
gtracer_environ_viewer_new(void)
{
  GTracerEnvironViewer * viewer;
  GtkWidget * vbox;
  GtkWidget * scrolled_window;
  GtkAdjustment * adjustment;
  GtkCellRenderer * text_renderer;
  GtkCellRenderer * editable_text_renderer;
  gint position;
  GtkTreeViewColumn * col;  
  GtkTreeModel  * model;
  GtkTreeSortable * sortable;
  GtkWidget * update_button;
  GtkWidget * add_button;
  GtkWidget * unset_button;

  
  viewer 		  = g_new(GTracerEnvironViewer, 1);

  model = gtracer_environ_model_new();
  sortable = GTK_TREE_SORTABLE(model);
  gtk_tree_sortable_set_sort_func(sortable, SORTID_PAIR,
				  gtracer_environ_model_comapre_func,
				  NULL, NULL);

  viewer->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_inspector_tag_to_widget (viewer->window);
  gtk_window_set_title(GTK_WINDOW(viewer->window),
		       _("Environment Variables"));
  g_signal_connect(G_OBJECT(viewer->window),
		   "delete_event",
		   G_CALLBACK(gtk_widget_hide_on_delete),
		   NULL);
  vbox = gtk_vbox_new(FALSE, 4);
  gtk_container_add(GTK_CONTAINER(viewer->window), vbox);
  
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  adjustment 	  = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));
  adjustment 	  = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));
  
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 2);

  viewer->view	= gtk_tree_view_new ();

  text_renderer = gtk_cell_renderer_text_new ();
  g_inspector_tag_to_object(G_OBJECT(text_renderer));

  position = 0;
  position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							position,
							_("Key"),
							text_renderer,
							gtracer_environ_model_extract_key,
							NULL,
							NULL);
  col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
  gtk_tree_view_column_set_sort_column_id(col, SORTID_PAIR);
  gtracer_register_tree_view_column(col);

  editable_text_renderer = gtk_cell_renderer_text_new ();
  g_inspector_tag_to_object(G_OBJECT(editable_text_renderer));

  g_object_set(editable_text_renderer, "editable", TRUE, NULL);
  position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							position,
							_("Value"),
							editable_text_renderer,
							gtracer_environ_model_extract_value,
							NULL,
							NULL);
  col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
  g_signal_connect(editable_text_renderer, 
		   "edited", 
		   G_CALLBACK(gtracer_environ_viewer_value_edited), 
		   model);
  gtracer_register_tree_view_column(col);

  gtk_tree_view_set_model (GTK_TREE_VIEW (viewer->view), model);
  g_object_unref (G_OBJECT(model));

  gtk_container_add(GTK_CONTAINER(scrolled_window), viewer->view);
    
  viewer->selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (viewer->view));
  g_inspector_tag_to_object(G_OBJECT(viewer->selection));
  gtk_tree_selection_set_mode (GTK_TREE_SELECTION (viewer->selection),
			       GTK_SELECTION_BROWSE);
  
  gtk_widget_set_size_request (viewer->view, 450, 250);

  viewer->statusbar = gtk_statusbar_new();
  gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(viewer->statusbar), FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), viewer->statusbar,
		      FALSE, FALSE, 2);

  update_button = gtk_button_new_with_label(_("Update"));
  g_signal_connect_swapped(G_OBJECT(update_button), "clicked",
			   G_CALLBACK(gtracer_environ_viewer_update),
			   viewer);
  gtk_box_pack_end(GTK_BOX(viewer->statusbar), update_button,
		   FALSE, FALSE, 2);

  add_button = gtk_button_new_with_label(_("Add"));
  g_signal_connect_swapped(G_OBJECT(add_button), "clicked",
			   G_CALLBACK(gtracer_environ_viewer_add),
			   viewer);
  gtk_box_pack_end(GTK_BOX(viewer->statusbar), add_button,
		   FALSE, FALSE, 2);
  
  unset_button = gtk_button_new_with_label(_("Unset"));
  g_signal_connect_swapped(G_OBJECT(unset_button), "clicked",
			   G_CALLBACK(gtracer_environ_viewer_unset),
			   viewer);
  gtk_box_pack_end(GTK_BOX(viewer->statusbar), unset_button,
		   FALSE, FALSE, 2);

  return viewer;
}

static gpointer
gtracer_environ_viewer_show(gpointer data, gpointer func_data, gpointer invoke_data)
{
  GTracerEnvironViewer * viewer = func_data;

  g_return_val_if_fail (viewer, NULL);
  g_return_val_if_fail (viewer->window, NULL);

  gtracer_environ_viewer_update(viewer);
  gtk_widget_show_all(viewer->window);
  
  return viewer->window;
}

static void
gtracer_environ_viewer_update(GTracerEnvironViewer * viewer)
{
  GtkTreeModel  * model;  
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(viewer->view));
  gtracer_environ_model_update(model);
}

static void
gtracer_environ_viewer_unset  (GTracerEnvironViewer * viewer)
{
  g_warning("Not implemented yet");
  /* TODO */
}

static void
gtracer_environ_viewer_add    (GTracerEnvironViewer * viewer)
{
  g_warning("Not implemented yet");
  /* TODO */
}

static void
gtracer_environ_viewer_value_edited(GtkCellRendererText * cell,
				    gchar               * path_string,
				    gchar               * new_text,
				    gpointer              user_data)
{
  GtkTreeModel  *model = user_data;
  GtkTreeIter iter;
  
  gtk_tree_model_get_iter_from_string(model, &iter, path_string);

  gtracer_environ_model_edit_value(model, &iter, new_text);
}

/* arch-tag: 062684f4-6dee-4ca0-af41-76bdc2bca9ae
   (do not change this comment) */
