/* fqarray.c --- Data structure to store the frequency of events

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "fqarray.h"
#include "bubble.h"
#include <stdio.h>

typedef struct _FqEntry   FqEntry;
typedef struct _DiffEntry DiffEntry;

struct _FqArray
{
  GArray *            array;
  FqArrayUpdateNotify notify;
  gpointer            client;
  GList *             diff_list;
};

struct _FqEntry
{
  gulong fq;
  GList* diff_list;
  gpointer user_data;  
};

struct _DiffEntry
{
  guint  index;
  gulong diff;
};

static gulong    fq_array_calc_diff(FqArray * fqa, guint index);
static void      fq_array_update_diff(FqArray * fqa, guint index);
static void      fq_array_decr_diff(FqArray * fqa, FqEntry * fqe);

#define          fq_array_get_larger(fqa,index) fq_array_get_entry((fqa), (index) - 1)
static void      fq_array_propagate_to_smaller(FqArray * fqa, guint index);

static guint     fq_array_seek(FqArray * fqa, guint index);
static FqEntry*  fq_array_get_entry(FqArray * fqa, guint index);
static void      fq_array_swap(FqArray * fqa, guint old_index, guint new_index);
static void      fq_array_notify(FqArray * fqa, guint index);

static inline guint fq_array_get_length (FqArray * fqa);

static gulong      fq_entry_diff(FqEntry * fqe);
static void        fq_entry_set_diff(FqEntry * fqe, gulong new_diff);
static void        fq_entry_set_index(FqEntry * fqe, guint new_index);


static DiffEntry * diff_entry_new(guint index, gulong diff);
static void        diff_entry_decr(DiffEntry * diff_entry);
static gint        diff_entry_compare(gconstpointer a, gconstpointer b);

static void      fq_entry_dump   (FqEntry * fqa);
static void      diff_entry_dump (DiffEntry * diff);

FqArray *
fq_array_new(FqArrayUpdateNotify notify,
	     gpointer client)
{
  FqArray * fqa;

  fqa 	     	 	= g_new(FqArray, 1);
  fqa->array 	 	= g_array_new(FALSE, FALSE, sizeof(FqEntry));
  fqa->notify 	 	= notify;
  fqa->client    	= client;
  fqa->diff_list        = NULL;
  
  return fqa;
}

guint
fq_array_add(FqArray * fqa, gpointer user_data)
{
  FqEntry fqe, *fqep;
  
  DiffEntry * diff_entry;

  guint index;
  gulong diff;
  
  g_return_val_if_fail (fqa, 0);

  fqe.fq    	 = 1;
  fqe.diff_list  = NULL;
  fqe.user_data  = user_data;

  fqa->array = g_array_append_val(fqa->array, fqe);
  
  index = fq_array_get_length(fqa) - 1;
  diff = fq_array_calc_diff(fqa, index);
  diff_entry = diff_entry_new(index, diff);
  fqa->diff_list = g_list_insert_sorted(fqa->diff_list,
					diff_entry,
					diff_entry_compare);

  fqep = fq_array_get_entry(fqa, index);
  fqep->diff_list = g_list_find(fqa->diff_list, 
				diff_entry);
  return index;
}

static gulong
fq_array_calc_diff(FqArray * fqa, guint index)
{
  g_assert(fqa);
  g_assert(index < fq_array_get_length(fqa));

  if (index == 0)
    return 0;
  else
    {
      FqEntry * fqe;
      FqEntry * larger_fqe;

      fqe = fq_array_get_entry(fqa, index);  
      larger_fqe = fq_array_get_larger(fqa, index);
      
      g_assert(larger_fqe->fq >= fqe->fq);
      
      return larger_fqe->fq - fqe->fq;
    }
}

static void
fq_array_update_diff(FqArray * fqa, guint index)
{
  gulong old_diff, new_diff;
  FqEntry * fqe;

  fqe = fq_array_get_entry(fqa, index); 
  old_diff = fq_entry_diff(fqe); 
  
  new_diff = fq_array_calc_diff(fqa, index);

  if (new_diff == old_diff) 
    return ;
  
  fq_entry_set_diff(fqe, new_diff);
  if (new_diff < old_diff)
    fqa->diff_list = g_list_sink(fqa->diff_list, 
				 fqe->diff_list, 
				 diff_entry_compare);
  else if (old_diff < new_diff)
    fqa->diff_list = g_list_surface(fqa->diff_list, 
				    fqe->diff_list, 
				    diff_entry_compare);
    
}

guint
fq_array_incr(FqArray * fqa, guint index)
{
  FqEntry * fqe;
  guint new_index;
  
  g_return_val_if_fail (fqa, 0);

  fqe = fq_array_get_entry(fqa, index);

  fqe->fq++;

  if (fq_entry_diff(fqe) > 0)
    {
      fq_array_decr_diff(fqa, fqe);
      fq_array_propagate_to_smaller(fqa, index);
      return index;
    }
  
  if (index == 0)
    {
      fq_array_propagate_to_smaller(fqa, index);
      return 0;
    }
  
  
  new_index = fq_array_seek(fqa, index);
  g_assert(new_index != index);

  fq_array_swap(fqa, index, new_index);

  fq_array_notify(fqa, index);
  fq_array_notify(fqa, new_index);

  return new_index;
}

static void
fq_array_decr_diff(FqArray * fqa, FqEntry * fqe)
{
  GList * list;
  
  list = fqe->diff_list;
  diff_entry_decr(list->data);
  fqa->diff_list = g_list_sink(fqa->diff_list, list, diff_entry_compare);
}

static FqEntry*
fq_array_get_entry(FqArray * fqa, guint index)
{
  g_return_val_if_fail  (index < fq_array_get_length(fqa), NULL);
  
  return &(g_array_index(fqa->array, FqEntry, index));
}

static guint
fq_array_seek(FqArray * fqa, guint index)
{
  guint i;
  guint new_index;
  FqEntry * fqe;

  g_assert(fqa);

  if (index == 0)
    return 0;

  new_index = index;
  for (i = index - 1; 0 < i; i--)
    {
      fqe = fq_array_get_entry(fqa, i);

      if (fq_entry_diff(fqe) != 0)
	return i;
    }

  return 0;
}

static void
fq_array_propagate_to_smaller(FqArray * fqa, guint index)
{
  g_assert(fqa);

  if ((index + 1) == fq_array_get_length(fqa))
    return ;

  fq_array_update_diff(fqa, index + 1);
}

static void
fq_array_swap(FqArray * fqa, guint old_index, guint new_index)
{
  FqEntry * new_fqe, *old_fqe;
  FqEntry tmp;
  
  g_assert(fqa);
  
  if (old_index == new_index) 
    return ;

  new_fqe  = fq_array_get_entry(fqa, new_index);
  old_fqe  = fq_array_get_entry(fqa, old_index);
  tmp 	   = *new_fqe;
  *new_fqe = *old_fqe;
  *old_fqe = tmp;

  fq_array_update_diff(fqa, new_index);
  fq_array_propagate_to_smaller(fqa, new_index);
  fq_entry_set_index(new_fqe, new_index);
    
  fq_array_update_diff(fqa, old_index);
  fq_array_propagate_to_smaller(fqa, old_index);
  fq_entry_set_index(old_fqe, old_index);
}

static void
fq_array_notify(FqArray * fqa, guint index)
{
  FqEntry * fqe;

  g_assert(fqa);
  g_assert(index < fq_array_get_length(fqa));

  if (fqa->notify == NULL) 
    return ;

  fqe = fq_array_get_entry(fqa, index);
  fqa->notify(fqa->client, index, fqe->user_data);
}

guint
fq_array_inflection_index(FqArray * fqa, gfloat rate)
{
  GList * list;
  DiffEntry * diff_entry;
  gulong target_diff;
  guint inflection_index = 0;
  
  g_return_val_if_fail (fqa, 0);
  g_return_val_if_fail (rate >  0.0, 0);
  g_return_val_if_fail (rate <= 1.0, 0);

  target_diff = (gfloat)(fq_array_frequency(fqa, 0)) * rate;
  
  
  list = fqa->diff_list;
  g_return_val_if_fail (list, 0);
  
  while (list) 
    {
      diff_entry       = list->data;
      
      if (diff_entry->index > inflection_index) 
	inflection_index = diff_entry->index;
      
      if (diff_entry->diff < target_diff)
	break;
      list = g_list_next(list);
    }
  
  return inflection_index;
}

/*
 * Accessor
 */ 
guint
fq_array_length(FqArray * fqa)
{
  return fq_array_get_length(fqa);
}

gulong
fq_array_frequency(FqArray * fqa, 
		   guint index)
{
  FqEntry * fqe;

  g_return_val_if_fail (index < fq_array_get_length(fqa), 0);

  fqe = fq_array_get_entry(fqa, index);
  return fqe->fq;
}

gulong
fq_array_diff(FqArray * fqa, guint index)
{
  FqEntry * fqe;
  g_return_val_if_fail (index < fq_array_get_length(fqa), 0);
  fqe = fq_array_get_entry(fqa, index);
  return fq_entry_diff(fqe);
}

gpointer
fq_array_user_data (FqArray * fqa, guint index)
{
  FqEntry * fqe;
  g_return_val_if_fail (index < fq_array_get_length(fqa), 0);
  fqe = fq_array_get_entry(fqa, index);
  return fqe->user_data;
}

static inline guint
fq_array_get_length (FqArray * fqa)
{
  g_return_val_if_fail (fqa, 0);
  return fqa->array->len;
}

static gulong
fq_entry_diff(FqEntry * fqe)
{
  DiffEntry * diff_entry;
  
  g_return_val_if_fail (fqe, 0);

  diff_entry = fqe->diff_list->data;
  return diff_entry->diff;

}

static void
fq_entry_set_diff(FqEntry * fqe, gulong new_diff)
{
  DiffEntry * diff_entry;
  
  g_return_if_fail (fqe);

  diff_entry = fqe->diff_list->data;
  diff_entry->diff = new_diff;
}

static void
fq_entry_set_index(FqEntry * fqe, guint new_index)
{
  DiffEntry * diff_entry;
  
  g_return_if_fail (fqe);

  diff_entry = fqe->diff_list->data;
  diff_entry->index = new_index;
}

static DiffEntry *
diff_entry_new(guint index, gulong diff)
{
  DiffEntry * entry;

  entry        = g_new(DiffEntry, 1);
  entry->index = index;
  entry->diff  = diff;

  return entry;
}

static void
diff_entry_decr(DiffEntry * diff_entry)
{
  diff_entry->diff--;
}

static gint
diff_entry_compare(gconstpointer a, gconstpointer b)
{
  const DiffEntry * entry_a = a;
  const DiffEntry * entry_b = b;

  if (entry_a->diff > entry_b->diff)
    return -1;
  else if (entry_a->diff < entry_b->diff)
    return 1;
  else
    return 0;
}

/*
 * Dump
 */
void
fq_array_dump(FqArray * fqa)
{
  GList * list;
  FqEntry * entry;
  guint length, i;
  length = fq_array_get_length(fqa);
  
  for(i = 0; i < length; i++)
    {
      entry = fq_array_get_entry(fqa, i);
      fprintf(stderr, "%u: ", i);
      fq_entry_dump(entry);
    }
 
  list = fqa->diff_list;
  while (list)
    {
      diff_entry_dump(list->data);
      list = list->next;
    }
}

static void
fq_entry_dump(FqEntry * fqa)
{
  fprintf(stderr, "fq=%lu, data=%s, ",
	  fqa->fq,
	  fqa->user_data);
  diff_entry_dump((DiffEntry *)(fqa->diff_list->data));
}

static void
diff_entry_dump(DiffEntry * diff)
{
  fprintf(stderr, "diff=%lu, index=%u\n", 
	  diff->diff,
	  diff->index);
}

/*
 * For testing
 */
/* gcc -O0 -g fqarray.c bubble.c -DFQ_ARRAY_TEST `pkg-config --libs --cflags glib-2.0`  -fprofile-arcs -ftest-coverage */
/* #define FQ_ARRAY_TEST */
#ifdef FQ_ARRAY_TEST

void
notify(gpointer client, guint new_index, gpointer user_data)
{
  g_message("%s[%u] = %lu, diff = %lu",
	    user_data, 
	    new_index,
	    fq_array_frequency(*(FqArray **)client, new_index),
	    fq_array_diff(*(FqArray **)client, new_index));
}

void
func(FqArray * fqa)
{
  g_message("---------------1-");
  fq_array_incr(fqa, 1);
  /* gdb[0]   = 2
     gcc[1]   = 1
     emacs[2] = 1
     -------------
     ** Message: gcc[1] = 1
     ** Message: gdb[0] = 2 */

  g_message("---------------2-");
  fq_array_incr(fqa, 1);
  /* gdb[0]   = 2
     gcc[1]   = 2
     emacs[2] = 1
     -------------
     {nothing} */
  
  g_message("---------------3-");
  fq_array_incr(fqa, 1);
  /* gcc[0]   = 3
     gdb[1]   = 2
     emacs[2] = 1
     -------------
     ** Message: gdb[1] = 2
     ** Message: gcc[0] = 3 */

  g_message("---------------4-");
  fq_array_incr(fqa, 2);
  /* gcc[0]   = 3
     gdb[1]   = 2
     emacs[2] = 2
     -------------
     {nothing} */

  g_message("---------------5-");
  fq_array_incr(fqa, 2);
  /* gcc[0]   = 3
     emacs[1] = 3
     gdb[2]   = 2
     -------------
     ** Message: gdb[2] = 2
     ** Message: emacs[1] = 3 */
  
  g_message("---------------6-");
  fq_array_incr(fqa, 1);
  /* emacs[0] = 4
     gcc[1]   = 3
     gdb[2]   = 2
     -------------
     ** Message: gcc[1] = 3
     ** Message: emacs[0] = 4 */
}

int
main(int argc, char ** argv)
{
  int i;
  FqArray * fqa;

  fqa = fq_array_new(notify, &fqa);

  fq_array_add(fqa, "gcc");
  fq_array_add(fqa, "gdb");
  fq_array_add(fqa, "emacs");

  /* gcc[0]   = 1
     gdb[1]   = 1
     emacs[2] = 1 */

  for(i = 0; i < 0xFFFF; i++)
    func(fqa);

  fq_array_incr(fqa, 0);  
  fq_array_incr(fqa, 0);  
  fq_array_incr(fqa, 0);
  
  fq_array_dump(fqa);

  return 0;
} 

#endif /* Def: FQ_ARRAY_TEST */

/* arch-tag: 2e79015e-9a7d-4875-898a-bca85d5ecb78
   (do not change this comment) */
