/* fqrecorder.c --- Data structure to store the frequency of events 
                    identified by strings

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "fqrecorder.h"
#include "fqarray.h"

struct _FqRecorder
{
  GHashTable * name_to_index;
  FqArray *    fq_array;
};

static void fq_recorder_notify(gpointer client,
			       guint new_index,
			       gpointer user_data);

FqRecorder *
fq_recorder_new           (void)
{
  FqRecorder * recorder;

  recorder = g_new(FqRecorder, 1);

  recorder->name_to_index = g_hash_table_new(g_str_hash,
					     g_str_equal);
  recorder->fq_array   = fq_array_new(fq_recorder_notify,
				      recorder->name_to_index);
  return recorder;
}

void
fq_recorder_record        (FqRecorder * recorder,
			   const gchar * string)
{
  gpointer tmp;
  guint index;

  g_return_if_fail (recorder);
  g_return_if_fail (string);
  
  
  tmp = g_hash_table_lookup(recorder->name_to_index,
			    string);

  if ((tmp 				     == NULL)
      && ((fq_array_length(recorder->fq_array) == 0)
	  || (!g_str_equal(fq_array_user_data(recorder->fq_array, 0),
			   string))))
    {
      gchar * key;
      key   = g_strdup(string);
      index = fq_array_add(recorder->fq_array, key);
      g_hash_table_insert(recorder->name_to_index,
			  key, 
			  GUINT_TO_POINTER(index));
    } 
  else 
    {
      index = GPOINTER_TO_UINT(tmp);
      fq_array_incr(recorder->fq_array, index);
    }
}

gboolean
fq_recorder_is_frequently (FqRecorder * recorder,
			   gfloat rate,
			   const gchar * string)
{
  gpointer tmp;
  guint index;
  guint inflection_index;
  
  tmp = g_hash_table_lookup(recorder->name_to_index,
			    string);
  if ((tmp == NULL)
      && ((fq_array_length(recorder->fq_array) == 0)
	  || (!g_str_equal(fq_array_user_data(recorder->fq_array, 0),
			   string))))
    return FALSE;

  index = GPOINTER_TO_UINT(tmp);
  inflection_index = fq_array_inflection_index(recorder->fq_array,
					       rate);
  
  if (index < inflection_index) 
    return TRUE;
  else
    return FALSE;

}

static void
fq_recorder_notify(gpointer client,
		   guint new_index,
		   gpointer user_data)
{
  GHashTable * name_to_index = client;
  
  g_hash_table_replace(name_to_index,
		       user_data,
		       GUINT_TO_POINTER(new_index));
}

void
fq_recorder_dump          (FqRecorder * fq_recorder)
{
  fq_array_dump(fq_recorder->fq_array);
}

/*
 * For testing
 */
/* gcc -O0 -g fqarray.c bubble.c fqrecorder.c -DFQ_RECORDER_TEST `pkg-config --libs --cflags glib-2.0`  -fprofile-arcs -ftest-coverage */
/* #define FQ_RECORDER_TEST */

#ifdef FQ_RECORDER_TEST

void
report(FqRecorder * fqr, const gchar * key)
{
  if (fq_recorder_is_frequently(fqr, key) == TRUE)
    g_message ("%s is frequently occurred.", key);
  else
    g_message ("%s is rarely occurred.", key);
}

int
main(int argc, char ** argv)
{
  FqRecorder * fqr;
  gchar * key;
  
  fqr = fq_recorder_new();
  
  fq_recorder_record(fqr, "gcc");
  fq_recorder_record(fqr, "emacs");
  fq_recorder_record(fqr, "gdb");
  fq_recorder_record(fqr, "emacs");
  fq_recorder_record(fqr, "gcc");
  fq_recorder_record(fqr, "emacs");
  fq_recorder_record(fqr, "emacs");
  fq_recorder_record(fqr, "emacs");
  fq_recorder_record(fqr, "gdb");
  fq_recorder_record(fqr, "gdb");
  
  fq_array_dump(fqr->fq_array);
  
  key = "emacs";
  report(fqr, key);
  key = "gdb";
  report(fqr, key);
  key = "gcc";
  report(fqr, key);
  
  return 0;
} 
#endif  /* FQ_RECORDER_TEST */

/* arch-tag: 86551c70-a176-4944-8c82-3f79f42d8c82
   (do not change this comment) */
