/* main.c --- gtracer main event loop

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include <gtk/gtk.h>
#include <gmodule.h>

#include <ginspector/ginspector-plugin-manager.h>
#include <ginspector/ginspector-preload-manager.h>

#include "i18n.h"
#include "plugin.h"
#include "service.h"

static guint    gtracer_idle_tag = 0;
static guint    gtracer_timeout_tag = 0;

static gboolean gtracer_main       (gpointer data);
static gboolean gtracer_reschedule (guint nlog);

static void     gtracer_start_tracing(void);
static gboolean gtracer_start_logging(gpointer unused);

/* Log constructor management */
static GData *  gtracer_log_constructors = NULL;
static gpointer gtracer_get_log_constructor  (const gchar * log_name);

GType GTracerLogType;
GTimeVal gtracer_epoch;

G_MODULE_EXPORT void
plugin_module_entry_point (void)
{
  {
    bindtextdomain (GETTEXT_PACKAGE, GTRACERLOCALEDIR);  
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  }
  {
    GTracerLogType = g_boxed_type_register_static(GTRACER_LOG_TYPE_NAME,
						  (GBoxedCopyFunc)gtracer_log_ref,
						  (GBoxedFreeFunc)gtracer_log_unref);
    g_datalist_init(&gtracer_log_constructors);
    gtracer_queue_init();
  }
  {
    static GInspectorPluginModuleInfo plugin_info;
    static GTracerPluginData plugin_data;
    static GTracerPluginOperations plugin_operations = {
      .push_log         = gtracer_queue_push_log,
      .get_log_constructor = gtracer_get_log_constructor,
    };
    GError * err = NULL;
    
    plugin_info.name 		 = GTRACER_PLUGIN_NAME;
    plugin_info.data 		 = &plugin_data;
    plugin_data.mops             = &plugin_operations;

    gtracer_viewers_init();
    gtracer_filters_init();
    gtracer_services_init(&err);
    if (err) 
      {
	/* TODO: Abort */
	g_message("Error in %s", __FUNCTION__);
      }
    g_inspector_plugin_register_module (&plugin_info);
  }
  {
    g_get_current_time(&gtracer_epoch);
    gtracer_start_tracing();
    /* Delay the queue activation, so the user can
       see gimp's splash window. */
    gtk_init_add(gtracer_start_logging, NULL);
  }
}

static gboolean
gtracer_start_logging(gpointer unused)
{
  gtracer_idle_tag = g_idle_add_full(G_PRIORITY_LOW,
				     gtracer_main,
				     NULL,
				     NULL);
  
  /* g_message("install: idle %u", gtracer_idle_tag); */

  /* When this function is called through gtk_init_add,
     the return value is not used.
     However, then this is called through g_timeout_add,
     it is removed from the source if it returns
     FALSE. */
  return FALSE;			
}

static gboolean
gtracer_main(gpointer unsed_data)
{
  GTracerLog * log;
  GTracerLogProtected * log_protected;
  guint nlog = 0;
  
  gtracer_queue_freeze();
  while ((log = gtracer_queue_pop_log()))
    {
      log_protected = (GTracerLogProtected * )log;
      gtracer_log_ref(log);
      
      if (!gtracer_log_is_nil(log_protected))
	{
	  GTracerFilter* filter;
	  GTracerViewer* viewer;

	  filter = gtracer_filters_find(log_protected->log_type);
	  if (gtracer_filter_filter_log(filter, log_protected, NULL) == FALSE)
	    goto filtered;
	  
	  viewer = gtracer_viewers_find(log_protected->log_type);
	  if (viewer == NULL)
	    goto no_viewer;

	  nlog++;
	  gtracer_viewer_push_log(viewer, ((GTracerLogProtected *)log), NULL);

	}
    filtered:
    no_viewer:
      gtracer_log_unref(log);
    }
  gtracer_queue_thaw();

  return gtracer_reschedule(nlog);
}

/**
 * gtracer_reschedule:
 * @Returns: 
 * @nlog: count of logs.
 * 
 * Reschedule the timing of gtracer main loop execution.
 **/
static gboolean
gtracer_reschedule   (guint nlog)
{
  static guint n_nolog        = 0;
  static int delay_array[] = {
    0,				/* 0, not used */
    50,				/* 1 */
    100,			/* 2 */
    200,			/* 3 */
    400,			/* 4 */
    800,			/* 5 */
#define delay_array_last_index 5
  };

  if (nlog == 0)
    {
      gint delay;
      gint index;
      
      n_nolog++;
      index = (n_nolog < delay_array_last_index)? n_nolog: delay_array_last_index;
      delay = delay_array[index];

      gtracer_timeout_tag = g_timeout_add(delay,
					 gtracer_start_logging,
					 NULL);
      /* g_message("delayed: timeout %u in %u", gtracer_timeout_tag, delay); */
      return FALSE;
    }
  else
    {
      /* g_message("cont: idle %u", gtracer_idle_tag); */
      n_nolog = 0;
      return TRUE;
    }
}

static void
gtracer_start_tracing(void)
{
  GInspectorPreloadModuleInfo * preload_info;
  GTracerPreloadData * preload_data;
  GTracerPreloadOperations * preload_operations;
  preload_info = g_inspector_preload_find_module(GTRACER_PRELOAD_NAME);
  g_return_if_fail (preload_info);

  preload_data = preload_info->data;
  g_return_if_fail (preload_data);
  
  preload_operations = preload_data->pops;
  g_return_if_fail (preload_operations);
  preload_operations->start_tracing();
}

/* ------------------------------------------------------------------
   Log constructor management
 ------------------------------------------------------------------ */
void
gtracer_register_log_constructor(GQuark log_type, gpointer log_constructor)
{
  g_datalist_id_set_data(&gtracer_log_constructors, log_type, log_constructor);
}
static gpointer
gtracer_get_log_constructor(const gchar * log_name)
{
  return g_datalist_get_data(&gtracer_log_constructors, log_name);
}

/* ------------------------------------------------------------------
   Object life cycle management 
   ------------------------------------------------------------------ */
void
gtracer_make_object_zombie(GObject * object)
{
  g_return_if_fail (object);
  g_object_set_data(object, GTRACER_OBJECT_ZOMBIE_KEY, (gpointer)TRUE);
}


/* ------------------------------------------------------------------
   GTracer Main Loop Information 
   Check functions and tags which are owned by the main loop.
   ------------------------------------------------------------------ */

/**
 * gtracer_check_function:
 * @Returns: TRUE if the tag is owned by gtracer main loop
 * @function: a function to be checked
 * 
 * 
 **/
gboolean
gtracer_check_function(gpointer function)
{
  return ( (function == gtracer_main) || (function == gtracer_start_logging) )
    ? TRUE
    : FALSE;
}

/**
 * gtracer_check_tag:
 * @Returns: TRUE if the tag is owned by gtracer main loop
 * @tag: a tag to be checked
 * 
 * 
 **/
gboolean
gtracer_check_tag(guint tag)
{
  return (tag == gtracer_idle_tag || tag == gtracer_timeout_tag)
    ? TRUE
    : FALSE;
}

/* arch-tag: 759cfa59-f805-4a3f-a90f-77182ec00d57
   (do not change this comment) */
