/* pmaps_service.c --- implementation of the memory mapping related service

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "pmaps_service.h"

#include "i18n.h"

#include <ginspector/ginspector-tag.h>
#include <pmaps.h>
#include <addrslv.h>
#include <fptreel.h>
#include <mainpos.h>
#include <string.h>

static PMaps * gtracer_pmaps = NULL;

GType AddrslvSymbolInfoType;
#define ADDRSLV_SYMBOLINFO_TYPE_NAME "AddrslvSymbolInfo"

static GType PMapsEntryType;
#define PMAPS_ENTRY_TYPE_NAME "PMapsEntry"

enum
{
  COL_ENTRY = 0,
  NUM_COLS
} ;

#define ITER_KEY           "gtracer_pmaps_model_iter"
#define ONLY_EXE_KEY       "gtracer_pmaps_model_exe"
#define LOGGING_KEY        "gtracer_pmaps_model_logging"

#define LOGGING_FILES_KEY  "gtracer_pmaps_model_logging_files" /* Error */

#define UPDATE_NOTIFY_KEY       "gtracer_pmaps_model_update_notify"
#define UPDATE_NOTIFY_DATA_KEY  "gtracer_pmaps_model_update_notify_data"

static void     gtracer_pmaps_fill_model        (PMapsEntry * entry, 
						 gpointer callback_data);
static void     gtracer_pmaps_set_logging       (PMapsEntry * entry, gpointer logging);
static gboolean gtracer_pmaps_get_logging       (PMapsEntry * entry);
static void     gtracer_pmaps_model_sync        (PMaps * pmaps, gboolean after, gpointer callback_data);

void
gtracer_pmaps_service_init (GError** err)
{
  GError * tmp_error = NULL;
  PMapsEntry * main_entry;
  
  gtracer_pmaps_log_class_init(&tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return ;
    }
  
  gtracer_pmaps = pmaps_new (getpid(), &tmp_error);
  if (tmp_error)
    {
      g_propagate_error (err, tmp_error);
      return;
    }

  AddrslvSymbolInfoType = g_boxed_type_register_static(ADDRSLV_SYMBOLINFO_TYPE_NAME,
						       (GBoxedCopyFunc)addrslv_symbolinfo_ref,
						       (GBoxedFreeFunc)addrslv_symbolinfo_unref);
  
  PMapsEntryType = g_boxed_type_register_static(PMAPS_ENTRY_TYPE_NAME,
						(GBoxedCopyFunc)pmaps_entry_ref,
						(GBoxedFreeFunc)pmaps_entry_unref);

  pmaps_foreach(gtracer_pmaps, gtracer_pmaps_set_logging, FALSE);
  main_entry = mainpos_find(gtracer_pmaps, NULL);
  
  if (main_entry) 
    gtracer_pmaps_set_logging(main_entry, (gpointer)TRUE);
  else
    g_warning("Cannot find main entry"); /* TODO */

  addrslv_init();
    
  gtracer_pmaps_viewer_init(&tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return ;
    }
}

AddrslvSymbolInfo *
gtracer_pmaps_address_resolve(guint address, GError ** err)
{
  gpointer bait = GUINT_TO_POINTER(address);
  guint fish;
  guint reel_stage;
  
  AddrslvSymbolInfo * symbol_info;

  GError * tmp_error = NULL;
  
  fish = fptreel_reelup_debug(bait, gtracer_pmaps,  &reel_stage, &tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return NULL;
    }
  
  symbol_info = addrslv_resolve(gtracer_pmaps, fish, &tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return NULL;
    }

  /* For DEBUG */
  if ((symbol_info == NULL) && (fish != address))
    {
      g_message("bait: %p", bait);
      g_message("fish: %x", fish);
      pmaps_dump(gtracer_pmaps, NULL);
      g_message("reel up stage: %u", reel_stage);
      g_message("pid: %u", getpid());
      while (1) sleep(1);
    }
  
  return symbol_info;
}

PMapsEntry*
gtracer_pmaps_address_find_entry (guint address, GError ** err)
{
  gpointer bait = GUINT_TO_POINTER(address);
  guint fish;
  guint reel_stage;
  
  GError * tmp_error = NULL;
  
  fish = fptreel_reelup_debug(bait, gtracer_pmaps,  &reel_stage, &tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return NULL;
    }

  return pmaps_find (gtracer_pmaps, fish);
}

gboolean
gtracer_pmaps_address_filter (guint address, GError ** err)
{
  gpointer bait = GUINT_TO_POINTER(address);
  guint fish;
  guint reel_stage;

  PMapsEntry * entry;
  gboolean logging;

  GError * tmp_error = NULL;
  
  fish = fptreel_reelup_debug(bait, gtracer_pmaps,  &reel_stage, &tmp_error);
  if (tmp_error) 
    {
      g_propagate_error(err, tmp_error);
      return FALSE;
    }

  entry = pmaps_find(gtracer_pmaps, fish);
  if (entry == NULL) 
    return FALSE;
  
  logging = gtracer_pmaps_get_logging(entry);
  if (logging == FALSE) 
    return FALSE;

  return TRUE;
}

void
gtracer_pmaps_update(GError ** err)
{
  pmaps_update(gtracer_pmaps, err);
}

static void
gtracer_pmaps_set_logging       (PMapsEntry * entry, gpointer logging)
{
  pmaps_entry_set_data(entry, LOGGING_KEY, logging, NULL);
}

static gboolean
gtracer_pmaps_get_logging       (PMapsEntry * entry)
{
  g_return_val_if_fail (entry, FALSE);
  return (gboolean)pmaps_entry_get_data(entry, LOGGING_KEY);
}

GtkTreeModel*   gtracer_pmaps_model_new(gboolean only_executable,
					GTracerPMapsModelUpdateNotify func,
					gpointer data)
{
  GtkListStore  *store;
  store = gtk_list_store_new (NUM_COLS, PMapsEntryType);
  g_inspector_tag_to_object(G_OBJECT(store));
  
  gtracer_pmaps_model_update(GTK_TREE_MODEL(store), only_executable);

  g_object_set_data(G_OBJECT(store), UPDATE_NOTIFY_KEY, func);
  g_object_set_data(G_OBJECT(store), UPDATE_NOTIFY_DATA_KEY, data);
  
  pmaps_add_update_notify(gtracer_pmaps, gtracer_pmaps_model_sync, store);

  return GTK_TREE_MODEL(store);
}

static void gtracer_pmaps_mode_gather_logging_files(PMapsEntry * entry, gpointer callback_data);
static void gtracer_pmaps_mode_return_logging_files(PMapsEntry * entry, gpointer callback_data);

static void
gtracer_pmaps_model_sync        (PMaps * pmaps, gboolean after, gpointer callback_data)
{
  GData * logging_files = NULL;

  if (after) {
    GtkTreeModel *model 	= callback_data;
    GTracerPMapsModelUpdateNotify func;
    gpointer data;

    logging_files = pmaps_get_data(pmaps, LOGGING_FILES_KEY);
    pmaps_foreach(pmaps, 
		  gtracer_pmaps_mode_return_logging_files, &logging_files);
    g_datalist_clear(&logging_files); 
    
    func = g_object_get_data(G_OBJECT(model), UPDATE_NOTIFY_KEY);
    data = g_object_get_data(G_OBJECT(model), UPDATE_NOTIFY_DATA_KEY);
    
    if (func) 
      func(model, data);
  } else {
    g_datalist_init(&logging_files);
    pmaps_foreach(pmaps, 
		  gtracer_pmaps_mode_gather_logging_files, &logging_files);
    pmaps_set_data(pmaps, LOGGING_FILES_KEY, logging_files, NULL);
  }
}

static void
gtracer_pmaps_mode_gather_logging_files(PMapsEntry * entry, gpointer callback_data)
{
  GData ** logging_files = callback_data;
  if ((entry->prot & PMAPS_EXECUTABLE)
      && gtracer_pmaps_get_logging(entry))
    g_datalist_set_data(logging_files, entry->filename, GINT_TO_POINTER(1));
}

static void
gtracer_pmaps_mode_return_logging_files(PMapsEntry * entry, gpointer callback_data)
{
  GData ** logging_files = callback_data;
  if (!(entry->prot & PMAPS_EXECUTABLE))
    return ;

  if (g_datalist_get_data(logging_files, entry->filename))
    gtracer_pmaps_set_logging(entry, (gpointer)TRUE);
}

void
gtracer_pmaps_model_update(GtkTreeModel* model, gboolean only_executable)
{
  GtkTreeIter iter;
  
  gtk_list_store_clear(GTK_LIST_STORE(model));
  
  g_object_set_data(G_OBJECT(model), ITER_KEY, &iter);
  g_object_set_data(G_OBJECT(model), ONLY_EXE_KEY, (gpointer)only_executable);
  pmaps_foreach(gtracer_pmaps, gtracer_pmaps_fill_model, model);
  g_object_set_data(G_OBJECT(model), ONLY_EXE_KEY, NULL);
  g_object_set_data(G_OBJECT(model), ITER_KEY, NULL);
}

static void
gtracer_pmaps_fill_model(PMapsEntry * entry, gpointer callback_data)
{
  GtkListStore* store 	   = callback_data;
  GtkTreeIter*  iter  	   = g_object_get_data(G_OBJECT(store), ITER_KEY);
  gboolean only_executable = (gboolean)g_object_get_data(G_OBJECT(store), ONLY_EXE_KEY);
  
  if ((!only_executable)
      || (only_executable && (entry->prot & PMAPS_EXECUTABLE)))
    {
      gtk_list_store_append (store, iter);
      gtk_list_store_set (store, iter, COL_ENTRY, entry, -1);
    }
}

void
gtracer_pmaps_model_extract_filename_data(GtkTreeViewColumn *col,
					  GtkCellRenderer   *renderer,
					  GtkTreeModel      *model,
					  GtkTreeIter       *iter,
					  gpointer           user_data)
{
  PMapsEntry * entry;
  gchar * slash_pos;
  gchar * text;
  gboolean short_filename = *(gboolean*)user_data;
  
  
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);

  if (!short_filename)
    {
      g_object_set(G_OBJECT(renderer), "text", entry->filename, NULL);
      return ;
    }

  slash_pos = strrchr(entry->filename, '/');

  if (slash_pos && slash_pos[1] != '\0') 
    {
      gchar * so_pos;
      
      text = g_strdup(slash_pos + 1);
      so_pos = g_strrstr(text, ".so");
      
      if (so_pos) 
	*(so_pos + 3) = '\0';
    }
  else
    text = g_strdup(entry->filename);

  g_object_set(G_OBJECT(renderer), "text", text, NULL);
  g_free(text);
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}

void
gtracer_pmaps_model_extract_prot_data     (GtkTreeViewColumn *col,
					   GtkCellRenderer   *renderer,
					   GtkTreeModel      *model,
					   GtkTreeIter       *iter,
					   gpointer           user_data)
{
  PMapsEntry * entry;
  gchar buffer[5];
  buffer[4] = '\0';
  
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);
  buffer[0] = (entry->prot & PMAPS_READABLE)? 'r': '-';
  buffer[1] = (entry->prot & PMAPS_WRITABLE)? 'w': '-';
  buffer[2] = (entry->prot & PMAPS_EXECUTABLE)? 'x': '-';
  buffer[3] = (entry->prot & PMAPS_SHARED)? '-': 'p';
  g_object_set(renderer, "text", buffer, NULL);
#undef BUFFER_LENGTH
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}

void
gtracer_pmaps_model_extract_start_data(GtkTreeViewColumn *col,
				       GtkCellRenderer   *renderer,
				       GtkTreeModel      *model,
				       GtkTreeIter       *iter,
				       gpointer           user_data)
{
  PMapsEntry * entry;
#define BUFFER_LENGTH 32
  gchar buffer[BUFFER_LENGTH];
  
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);
  snprintf(buffer, BUFFER_LENGTH, "%08x", entry->start);
  g_object_set(renderer, "text", buffer, NULL);
#undef BUFFER_LENGTH
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}


void
gtracer_pmaps_model_extract_end_data(GtkTreeViewColumn *col,
				     GtkCellRenderer   *renderer,
				     GtkTreeModel      *model,
				     GtkTreeIter       *iter,
				     gpointer           user_data)
{
  PMapsEntry * entry;
#define BUFFER_LENGTH 32
  gchar buffer[BUFFER_LENGTH];
  
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);
  snprintf(buffer, BUFFER_LENGTH, "%08x", entry->end);
  g_object_set(renderer, "text", buffer, NULL);
#undef BUFFER_LENGTH
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}

void
gtracer_pmaps_model_extract_logging_data    (GtkTreeViewColumn *col,
					     GtkCellRenderer   *renderer,
					     GtkTreeModel      *model,
					     GtkTreeIter       *iter,
					     gpointer           user_data)
{
  PMapsEntry * entry;
  gboolean logging;
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);
  if (entry->prot & PMAPS_EXECUTABLE) 
    {
      logging = gtracer_pmaps_get_logging(entry);
      g_object_set(renderer, "active", logging, NULL); 
    } 
  else 
    g_object_set(renderer, "visible", FALSE, NULL); 
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}

void
gtracer_pmaps_model_toggle_logging_data (GtkTreeModel  *model,
					 GtkTreeIter* iter)
{
  PMapsEntry * entry;
  gboolean logging;
  gtk_tree_model_get(model, iter, COL_ENTRY, &entry, -1);
  logging = gtracer_pmaps_get_logging(entry);
  gtracer_pmaps_set_logging(entry, (gpointer)(!logging));
  g_boxed_free(PMapsEntryType, entry);/* MODEL_GET_FREE */
}

gchar *
gtracer_pmaps_message_new        (AddrslvSymbolInfo * symbol_info,
				  guint address)
{
  PMapsEntry * entry;
  gchar * msg;
  GError * err = NULL;
  
  if (symbol_info
      && (symbol_info->srcfile_quark != (GQuark)0)
      && (symbol_info->lineno 	     != 0))
    return g_strdup_printf(_("File: %s, Line: %u"),
			   g_quark_to_string(symbol_info->srcfile_quark),
			   symbol_info->lineno);
  
  entry = gtracer_pmaps_address_find_entry(address, 
					   &err);
  if (err)
    {
      msg = g_strdup_printf(_("Error[entry]: %s"), err->message);
      g_clear_error(&err);
      return msg;
    }
  if (entry)
    return g_strdup_printf(_("Library: %s, Region: %p - %p"),
			   entry->filename, 
			   GUINT_TO_POINTER(entry->start), 
			   GUINT_TO_POINTER(entry->end));
  return g_strdup(_("No extra information"));
}

/* arch-tag: 26e8bb6f-7bc8-4a4a-9560-c164c453399f
   (do not change this comment) */
