/* pmaps_viewer.c --- process memory mapping viewer

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "pmaps_service.h"
#include "plugin.h"
#include "utils.h"

#include "i18n.h"

#include <gtk/gtk.h>

#include <ginspector/ginspector-core.h>
#include <ginspector/ginspector-tag.h>
#include <ginspector/ginspector-gtk-tag.h>

static void gtracer_pmaps_viewer_func (GTracerLogProtected * log,
				       gpointer viewer_data,
				       GTracerProxyLog * proxy_log);

typedef struct _GTracerPMapsViewer GTracerPMapsViewer;
struct _GTracerPMapsViewer
{
  GtkWidget * window;
  GtkWidget * view;
  gboolean    only_executable;
  gboolean    protection;
  gboolean    short_filename;
  GtkWidget * statusbar;
};

static GTracerPMapsViewer * gtracer_pmaps_viewer_new  (void);
static gpointer             gtracer_pmaps_viewer_show (gpointer data, 
						       gpointer func_data, 
						       gpointer invoke_data);
static void                 gtracer_pmaps_viewer_toggle_logging(GtkCellRendererToggle *cell,
								gchar *path_string,				      
								gpointer user_data);

static void                 gtracer_pmaps_viewer_update(GtkTreeModel* model,
							gpointer user_data);

static void                 gtracer_pmaps_viewer_clicked_update(GtkButton * button,
								gpointer user_data);

void
gtracer_pmaps_viewer_init (GError ** err)
{
  GTracerPMapsViewer * viewer;
  
  viewer = gtracer_pmaps_viewer_new();
  gtracer_viewers_install(GTRACER_PMAPS_LOG_NAME,
			  gtracer_pmaps_viewer_func,
			  NULL,
			  viewer);

  g_inspector_register_procedure ("pmaps-viewer",
				  "Pmaps Viewer",
				  G_INSPECTOR_PROCEDURE_TYPE_WINDOW,
				  G_INSPECTOR_ELEMENT_TYPE_NONE,
				  G_TYPE_NONE,
				  gtracer_pmaps_viewer_show,
				  viewer);
  g_tracer_append_menu_item(_("/View/Memory Mapping"), 
			    NULL, 
			    "pmaps-viewer");
}

static gpointer
gtracer_pmaps_viewer_show(gpointer data, gpointer func_data, gpointer invoke_data)
{
  GTracerPMapsViewer * viewer = func_data;

  g_return_val_if_fail (viewer, NULL);
  g_return_val_if_fail (viewer->window, NULL);
  gtk_widget_show_all(viewer->window);
  return viewer->window;
}

static GTracerPMapsViewer *
gtracer_pmaps_viewer_new(void)
{
  GTracerPMapsViewer * viewer;
  GtkWidget          * vbox;
  GtkWidget          * hbox;
  GtkWidget          * sep;
  GtkWidget          * scrolled_window;
  GtkAdjustment      * adjustment;
  
  viewer 		  = g_new(GTracerPMapsViewer, 1);

  viewer->only_executable = TRUE;
  viewer->protection      = TRUE;
  viewer->short_filename  = TRUE;
  
  viewer->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_inspector_tag_to_widget (viewer->window);
  gtk_window_set_title(GTK_WINDOW(viewer->window),
		       _("Process Memory Mapping"));
  g_signal_connect(G_OBJECT(viewer->window),
		   "delete_event",
		   G_CALLBACK(gtk_widget_hide_on_delete),
		   NULL);

  
  vbox = gtk_vbox_new (FALSE, /* !homogeneous */
		       4);
  gtk_container_add(GTK_CONTAINER(viewer->window),
		    vbox);
  
  /*
   * Controllers
   */
  {
    GtkWidget          * executable_check;
    GtkWidget          * protection_check;
    GtkWidget          * short_filename_check;
  
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_box_pack_start (GTK_BOX (vbox), hbox,
			FALSE, FALSE, 2);

    executable_check = gtk_check_button_new_with_label(_("Showing Only Executable"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(executable_check),
				 viewer->only_executable);
    gtk_widget_set_sensitive(executable_check, FALSE); /* TILL GVAIRABLE */
    gtk_box_pack_start(GTK_BOX(hbox), executable_check,
		       FALSE, FALSE, 2);

    protection_check = gtk_check_button_new_with_label(_("Showing Protection"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(protection_check),
				 viewer->protection);
    gtk_widget_set_sensitive(protection_check, FALSE); /* TILL GVAIRABLE */
    gtk_box_pack_start(GTK_BOX(hbox), protection_check,
		       FALSE, FALSE, 2);
    
    short_filename_check = gtk_check_button_new_with_label(_("Using Short Filename"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(short_filename_check),
				 viewer->short_filename);
    gtk_widget_set_sensitive(short_filename_check, FALSE); /* TILL GVAIRABLE */
    gtk_box_pack_start(GTK_BOX(hbox), short_filename_check,
		       FALSE, FALSE, 2);

  }

  sep = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), sep,
		      FALSE, FALSE, 0);
  
  /*
   * Tree View
   */
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  adjustment 	  = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));
  adjustment 	  = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  g_inspector_tag_to_object(G_OBJECT(adjustment));
  
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (vbox), scrolled_window,
		      TRUE, TRUE, 2);


  {
    GtkCellRenderer    * text_renderer;
    GtkCellRenderer    * toggle_renderer;
    GtkTreeModel  *model;
    gint position;
    GtkTreeViewColumn * col;
    GtkTreeSelection   * list_selection;
    
    viewer->view	= gtk_tree_view_new ();
    model = gtracer_pmaps_model_new(viewer->only_executable,
				    gtracer_pmaps_viewer_update,
				    viewer);

    text_renderer = gtk_cell_renderer_text_new ();
    g_inspector_tag_to_object(G_OBJECT(text_renderer));

    position = 0;
    position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							  position,
							  _("Start"),
							  text_renderer,
							  gtracer_pmaps_model_extract_start_data,
							  NULL,
							  NULL);
    col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
    gtracer_register_tree_view_column(col);

    position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							  position,
							  _("End"),
							  text_renderer,
							  gtracer_pmaps_model_extract_end_data,
							  NULL,
							  NULL);
    col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
    gtracer_register_tree_view_column(col);
    
    position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							  position,
							  _("Mapped File"),
							  text_renderer,
							  gtracer_pmaps_model_extract_filename_data,
							  &(viewer->short_filename),
							  NULL);
    col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
    gtracer_register_tree_view_column(col);
    
    position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							  position,
							  _("Protection"),
							  text_renderer,
							  gtracer_pmaps_model_extract_prot_data,
							  NULL,
							  NULL);
    col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
    gtracer_register_tree_view_column(col);

    toggle_renderer = gtk_cell_renderer_toggle_new();
    g_inspector_tag_to_object(G_OBJECT(toggle_renderer));
    g_signal_connect(toggle_renderer, "toggled", 
		     G_CALLBACK(gtracer_pmaps_viewer_toggle_logging), model);
    g_object_set(toggle_renderer, "activatable", TRUE, NULL);
    position = gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(viewer->view),
							  position,
							  _("Logging"),
							  toggle_renderer,
							  gtracer_pmaps_model_extract_logging_data,
							  NULL,
							  NULL);
    col = gtk_tree_view_get_column(GTK_TREE_VIEW(viewer->view), position - 1);
    gtracer_register_tree_view_column(col);
    
    gtk_tree_view_set_model (GTK_TREE_VIEW (viewer->view), model);
    g_object_unref (model);

    gtk_container_add(GTK_CONTAINER(scrolled_window),
		      viewer->view);
    
    list_selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (viewer->view));
    g_inspector_tag_to_object(G_OBJECT(list_selection));

    gtk_widget_set_size_request (viewer->view, -1, 200);
  }

  /*
   * Statusbar
   */
  {
    GtkWidget          * update_button;
    
    viewer->statusbar = gtk_statusbar_new();
    gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(viewer->statusbar), FALSE);
    gtk_box_pack_start (GTK_BOX (vbox), viewer->statusbar,
			FALSE, FALSE, 2);
    update_button = gtk_button_new_with_label(_("Update"));
    g_signal_connect(G_OBJECT(update_button), "clicked",
		     G_CALLBACK(gtracer_pmaps_viewer_clicked_update),
		     viewer);
    gtk_box_pack_end(GTK_BOX(viewer->statusbar), update_button,
		     FALSE, FALSE, 2);
  }

  return viewer;
}

static void
gtracer_pmaps_viewer_toggle_logging(GtkCellRendererToggle *cell,
				    gchar *path_string,				      
				    gpointer user_data)
{
  GtkTreeModel  *model = user_data;
  GtkTreeIter iter;
  
  gtk_tree_model_get_iter_from_string(model, &iter, path_string);
  gtracer_pmaps_model_toggle_logging_data(model, &iter);
}

static void
gtracer_pmaps_viewer_func(GTracerLogProtected * log,
			  gpointer viewer_data,
			  GTracerProxyLog * proxy_log)
{
  GError * err = NULL;
  gtracer_pmaps_update(&err);
  /* TODO: err => statusbar */
 
  /* DEBUG */
  /* gtracer_log_dump(log); */
}

static void
gtracer_pmaps_viewer_update(GtkTreeModel* model, gpointer user_data)
{
  GTracerPMapsViewer * viewer = user_data;

  gtracer_pmaps_model_update(model, viewer->only_executable);
}

static void
gtracer_pmaps_viewer_clicked_update(GtkButton * button, gpointer user_data)
{
  GError * err = NULL;
  gtracer_pmaps_update(&err);
  /* TODO: err => statusbar */
}

/* arch-tag: afb0c11a-e2b8-4f50-a5b7-0d620bf3bc3b
   (do not change this comment) */
