/* signal_log.c --- log class for recording something event about signals

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include <glib.h>
#include <gtk/gtk.h>

#include <string.h>
#include <addrslv.h>

#include <ginspector/ginspector-tag.h>
#include <ginspector/ginspector-gtk-tag.h>
#include "gtracer.h"
#include "plugin.h"
#include "service.h"
#include "signal_service.h"


static GQuark gtracer_callback_log_type = (GQuark)0;
struct _GTracerCallbackLog
{
  GTracerSignalLog       log_signal;
  gchar*                 signal;
  GCallback              handler;
  gboolean               after;
};

static GTracerLog* gtracer_callback_log_new          (gpointer instance, 
						      const gchar * signal, 
						      GCallback handler, 
						      gboolean after);
static void         gtracer_callback_log_dump         (GTracerLogProtected * log_protected);
static void         gtracer_callback_log_free         (GTracerLogProtected * log_protected);
static gchar*       gtracer_callback_log_create_notes (GTracerSignalLog * signal_log);
static gchar*       gtracer_callback_log_create_message (GTracerSignalLog * signal_log);
static const gchar* gtracer_callback_log_extract_signal (GTracerSignalLog * log);

static GQuark gtracer_connect_log_type = (GQuark)0;
struct _GTracerConnectLog
{
  GTracerSignalLog    log_signal;
  gchar*              signal;
  GCallback           handler;
  GConnectFlags       connect_flags;
  gulong              handler_id;
};

static GTracerLog*  gtracer_connect_log_new  (gpointer instance, 
					      const gchar * signal, 
					      GCallback handler, 
					      GConnectFlags  connect_flags, 
					      gulong handler_id,
					      GSList * backtrace);
static void         gtracer_connect_log_dump (GTracerLogProtected * log_protected);
static void         gtracer_connect_log_free (GTracerLogProtected * log_protected);
static gchar*       gtracer_connect_log_create_notes (GTracerSignalLog * signal_log);
static gchar*       gtracer_connect_log_create_message (GTracerSignalLog * signal_log);
static const gchar* gtracer_connect_log_extract_signal (GTracerSignalLog * log);

/* Connect Stock Handling  */
typedef struct _GTracerConnectStock GTracerConnectStock;
struct _GTracerConnectStock
{
  GCallback           handler;
  GConnectFlags       connect_flags;
  gchar               signal[0];
};

static GTracerConnectStock * gtracer_instance_steal_stock(gpointer instance, 
							  gulong handler_id);
static GTracerConnectStock * gtracer_instance_get_stock(gpointer instance, 
							gulong handler_id);


static GQuark gtracer_disconnect_log_type = (GQuark)0;
struct _GTracerDisconnectLog
{
  GTracerSignalLog    log_signal;
  gulong              handler_id;
};

static GTracerLog*           gtracer_disconnect_log_new            (gpointer instance, gulong handler_id, 
								    GSList * backtrace);
static void                  gtracer_disconnect_log_dump           (GTracerLogProtected * log_protected);
static void                  gtracer_disconnect_log_free           (GTracerLogProtected * log_protected);
static gchar*                gtracer_disconnect_log_create_notes   (GTracerSignalLog * signal_log);
static gchar*                gtracer_disconnect_log_create_message (GTracerSignalLog * signal_log);
static GTracerConnectStock * gtracer_disconnect_log_get_stock      (GTracerDisconnectLog * disconnect_log);
static const gchar*          gtracer_disconnect_log_extract_signal (GTracerSignalLog * log);

static GQuark gtracer_emit_log_type = (GQuark)0;
struct _GTracerEmitLog
{
  GTracerSignalLog    log_signal;
  guint               signal_id;
  GQuark              quark;
  gboolean            after;
};

static GTracerLog*  gtracer_emit_log_new          (gpointer instance, 
						   guint signal_id, 
						   GQuark quark, 
						   gboolean after,
						   GSList * backtrace);
static void         gtracer_emit_log_dump         (GTracerLogProtected * log_protected);
static void         gtracer_emit_log_free         (GTracerLogProtected * log_protected);
static gchar*       gtracer_emit_log_create_notes (GTracerSignalLog * signal_log);
static const gchar* gtracer_emit_log_extract_signal (GTracerSignalLog * log);

/* ------------------------------------------------------------------
   Subststem initializer
   ------------------------------------------------------------------ */
void
gtracer_signal_log_class_init    (GError ** err)
{
  gtracer_callback_log_type = g_quark_from_static_string(GTRACER_CALLBACK_LOG_NAME);
  gtracer_register_log_constructor(gtracer_callback_log_type,
				   gtracer_callback_log_new);
  gtracer_connect_log_type = g_quark_from_static_string(GTRACER_CONNECT_LOG_NAME);
  gtracer_register_log_constructor(gtracer_connect_log_type,
				   gtracer_connect_log_new);
  gtracer_disconnect_log_type = g_quark_from_static_string(GTRACER_DISCONNECT_LOG_NAME);
  gtracer_register_log_constructor(gtracer_disconnect_log_type,
				   gtracer_disconnect_log_new);

  gtracer_emit_log_type = g_quark_from_static_string(GTRACER_EMIT_LOG_NAME);
  gtracer_register_log_constructor(gtracer_emit_log_type,
				   gtracer_emit_log_new);
}


/* ------------------------------------------------------------------
   Callback
   ------------------------------------------------------------------ */
gpointer
gtracer_callback_log_handler(GTracerCallbackLog * log)
{
  g_return_val_if_fail (log, NULL);
  return log->handler;
}

gboolean
gtracer_callback_log_after  (GTracerCallbackLog * log)
{
  g_return_val_if_fail (log, FALSE);
  return log->after;
}

static GTracerLog *
gtracer_callback_log_new(gpointer instance, const char * signal, GCallback handler, gboolean after)
{
  GTracerCallbackLog *cb_log;
  static GTracerLogProtectedOperations callback_props = {
    .dump       = gtracer_callback_log_dump,
    .free       = gtracer_callback_log_free,
  };
  static GTracerSignalLogOperations callback_slops = {
    .extract_signal = gtracer_callback_log_extract_signal,
    .create_notes   = gtracer_callback_log_create_notes,
    .create_message = gtracer_callback_log_create_message,
  };

  g_return_val_if_fail (signal, NULL);
  g_return_val_if_fail (handler, NULL);

  if (g_inspector_tag_check_object (instance))
    return gtracer_nil_log();
  if (GTK_IS_WIDGET(instance)
      && g_inspector_tag_check_object (instance))
    return gtracer_nil_log();

  cb_log 	   = g_new(GTracerCallbackLog, 1);
  gtracer_signal_log_init((GTracerSignalLog *)cb_log, 
			  gtracer_callback_log_type, 
			  &callback_props,
			  instance,
			  NULL,
			  &callback_slops);
  cb_log->signal   = g_strdup(signal);
  g_strdelimit (cb_log->signal, G_STR_DELIMITERS ":^", '-');

  cb_log->handler  = handler;
  cb_log->after    = after;
  
  return (GTracerLog *)cb_log;
}

static void
gtracer_callback_log_dump  (GTracerLogProtected * log_protected)
{
  GTracerSignalLog * log_signal = (GTracerSignalLog *)log_protected;
  GTracerCallbackLog * cb_log = (GTracerCallbackLog *)log_signal;
  AddrslvSymbolInfo * symbol_info;
  gpointer handler;
  GError * err = NULL;
  
  g_return_if_fail (cb_log);

  handler = gtracer_callback_log_handler(cb_log);
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler), 
					      &err);
  if (err) 
    {
      g_message("%s", err->message);
      g_clear_error(&err);
    }
  else if (symbol_info)
    {
      g_message ("<%s class=\"%s\" address=\"%p\" signal=\"%s\" function=\"%s\" timing=\"%s\"/>", 
		 GTRACER_CALLBACK_LOG_NAME,
		 gtracer_signal_log_class_name(log_signal),
		 gtracer_signal_log_instance(log_signal),
		 gtracer_signal_log_signal(log_signal),
		 (symbol_info->funcname)? (symbol_info->funcname): "unknown",
		 gtracer_callback_log_after(cb_log)?"after":"before");
    }
}

static void
gtracer_callback_log_free    (GTracerLogProtected * log_protected)
{
  GTracerCallbackLog * cb_log = (GTracerCallbackLog *)log_protected;

  gtracer_signal_log_fin((GTracerSignalLog *)log_protected);
  g_free(cb_log->signal);
  g_free(cb_log);
}

static gchar*
gtracer_callback_log_create_notes(GTracerSignalLog * signal_log)
{
  GTracerCallbackLog * cb_log = (GTracerCallbackLog *)signal_log;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;
  GError* err = NULL;
  
  g_return_val_if_fail(cb_log, NULL);
  handler     = gtracer_callback_log_handler(cb_log);
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {
      /* TODO */
      g_message ("%s", err->message);
      g_clear_error(&err);
      return g_strdup_printf(" => %p", handler);
    }
  
  if (symbol_info == NULL)
    return g_strdup_printf(" => %p", handler);

  if (symbol_info->funcname == NULL) 
    return g_strdup_printf(" => %p", handler);
			   
  
  return g_strdup_printf(" => %s()", symbol_info->funcname);
}

static gchar*
gtracer_callback_log_create_message (GTracerSignalLog * signal_log)
{
  GTracerCallbackLog * cb_log = (GTracerCallbackLog *)signal_log;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;
  gchar * msg;
  GError* err = NULL;
  
  
  g_return_val_if_fail(cb_log, NULL);

  handler = gtracer_callback_log_handler(cb_log);
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {
      msg = g_strdup_printf("Error: %s", err->message);
      g_clear_error(&err);
      return msg;
    }

  return gtracer_pmaps_message_new(symbol_info, GPOINTER_TO_UINT(handler));
}

static const gchar*
gtracer_callback_log_extract_signal(GTracerSignalLog * signal_log)
{
  GTracerCallbackLog * log = (GTracerCallbackLog *)signal_log;
  g_return_val_if_fail (log, NULL);
  return log->signal;
}


/* ------------------------------------------------------------------
   Connect
   ------------------------------------------------------------------ */
gpointer
gtracer_connect_log_handler(GTracerConnectLog * log)
{
  g_return_val_if_fail (log, NULL);
  return log->handler;
}

GConnectFlags
gtracer_connect_log_flags         (GTracerConnectLog * log)
{
  g_return_val_if_fail (log, (GConnectFlags)0);
  return log->connect_flags;
}

gulong
gtracer_connect_log_handler_id    (GTracerConnectLog * log)
{
  g_return_val_if_fail (log, 0);
  return log->handler_id;
}

static GTracerLog*
gtracer_connect_log_new  (gpointer instance, const gchar * signal, GCallback handler, 
			  GConnectFlags  connect_flags, gulong handler_id, GSList * backtrace)
{
  GTracerConnectLog *conn_log;
  static GTracerLogProtectedOperations connect_props = {
    .dump       = gtracer_connect_log_dump,
    .free       = gtracer_connect_log_free,
  };
  static GTracerSignalLogOperations connect_slops = {
    .extract_signal = gtracer_connect_log_extract_signal,
    .create_notes   = gtracer_connect_log_create_notes,
    .create_message = gtracer_connect_log_create_message,
  };
  g_return_val_if_fail (signal, NULL);
  g_return_val_if_fail (handler, NULL);
  if (g_inspector_tag_check_object (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }
  if (GTK_IS_WIDGET(instance)
      && g_inspector_tag_check_widget (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }

  conn_log 	     	  = g_new(GTracerConnectLog, 1);
  gtracer_signal_log_init((GTracerSignalLog *)conn_log, 
			  gtracer_connect_log_type, 
			  &connect_props,
			  instance,
			  backtrace,
			  &connect_slops);

  conn_log->signal     	  = g_strdup(signal);
  g_strdelimit (conn_log->signal, G_STR_DELIMITERS ":^", '-');
  
  conn_log->handler    	  = handler;
  conn_log->connect_flags = connect_flags;
  conn_log->handler_id 	  = handler_id;

  return (GTracerLog *)conn_log;
}

static void
gtracer_connect_log_dump (GTracerLogProtected * log_protected)
{
  GTracerSignalLog * log_signal = (GTracerSignalLog *)log_protected;
  GTracerConnectLog * conn_log = (GTracerConnectLog *)log_signal;
  
  g_return_if_fail (conn_log);
  g_message ("<%s class=\"%s\" address=\"%p\" signal=\"%s\" handler_id=\"%lu\"/>", 
	     GTRACER_CONNECT_LOG_NAME, 
	     gtracer_signal_log_class_name(log_signal),
	     gtracer_signal_log_class_name(log_signal),
	     conn_log->signal,
	     conn_log->handler_id);
}

static void
gtracer_connect_log_free (GTracerLogProtected * log_protected)
{
  GTracerConnectLog * conn_log = (GTracerConnectLog *)log_protected;

  gtracer_signal_log_fin((GTracerSignalLog *)log_protected);
  g_free(conn_log->signal);
  g_free(conn_log);
}

static gchar*
gtracer_connect_log_create_notes(GTracerSignalLog * signal_log)
{
  GTracerConnectLog * connect_log = (GTracerConnectLog *)signal_log;
  gpointer handler;
  gulong handler_id;
  AddrslvSymbolInfo * symbol_info;
  GError* err = NULL;
  
  g_return_val_if_fail(connect_log, NULL);
  handler     = gtracer_connect_log_handler(connect_log);
  handler_id  = gtracer_connect_log_handler_id(connect_log);
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {
      /* TODO */
      g_message ("%s", err->message);
      g_clear_error(&err);
      return g_strdup_printf(" + %p # %lu", 
			     handler,
			     handler_id);
    }

  if (symbol_info == NULL)
    return g_strdup_printf(" + %p # %lu", 
			   handler,
			   handler_id);

  if (symbol_info->funcname == NULL) 
    return g_strdup_printf(" + %p # %lu", 
			   handler,
			   handler_id);
			   
  
  return g_strdup_printf(" + %s() # %lu", 
			 symbol_info->funcname,
			 handler_id);
}

static gchar*
gtracer_connect_log_create_message (GTracerSignalLog * signal_log)
{
  GTracerConnectLog * conn_log = (GTracerConnectLog *)signal_log;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;
  gchar * msg;
  GError* err = NULL;
  
  g_return_val_if_fail(conn_log, NULL);

  handler = gtracer_connect_log_handler(conn_log);
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {
      msg = g_strdup_printf("Error: %s", err->message);
      g_clear_error(&err);
      return msg;
    }

  return gtracer_pmaps_message_new(symbol_info, GPOINTER_TO_UINT(handler));
}

static const gchar*
gtracer_connect_log_extract_signal (GTracerSignalLog * log)
{
  g_return_val_if_fail (log, NULL);
  return ((GTracerConnectLog *)log)->signal;
}

/* 
 * Connect Stock Handling 
 */
#define GTRACER_CONNECT_STOCK_KEY "gtracer_connect_stock_key"
static GTracerConnectStock * gtracer_connect_log_stock_new(gpointer handler,
							   GConnectFlags connect_flags,
							   const gchar * signal);
static void                  gtracer_connect_log_stock_free(GTracerConnectStock * stock);

static GHashTable *          gtracer_instance_get_stock_table(gpointer instance, 
							      gboolean create_if_no_table);

void
gtracer_connect_log_install_stock (GTracerConnectLog * log)
{
  gpointer instance;
  gulong   handler_id;
  gpointer handler;
  const gchar* signal;
  GConnectFlags connect_flags;
  
  GTracerConnectStock * stock;
  GHashTable* stock_table;
  
  g_return_if_fail (log);

  instance 	= gtracer_signal_log_instance((GTracerSignalLog *)log);

  stock_table = gtracer_instance_get_stock_table(instance, TRUE);

  handler_id    = gtracer_connect_log_handler_id(log); 
  handler       = gtracer_connect_log_handler(log);
  signal   	= gtracer_signal_log_signal((GTracerSignalLog *)log);
  connect_flags = gtracer_connect_log_flags(log);

  
  stock = gtracer_connect_log_stock_new(handler, 
					connect_flags, 
					signal);
  
  /* TODO: gulong it too large to convert pointer.*/
  g_hash_table_insert(stock_table, 
		      GUINT_TO_POINTER(handler_id), 
		      stock);
}

static GTracerConnectStock *
gtracer_connect_log_stock_new(gpointer handler,
			      GConnectFlags connect_flags,
			      const gchar * signal)
{
  GTracerConnectStock * stock;
  gulong alloc_bytes;
  gulong signal_length;
  
  signal_length = strlen(signal);
  alloc_bytes = sizeof(GTracerConnectStock) 
    + signal_length 
    + 1				/* For '\0' */
    ;
  stock 	       = g_malloc(alloc_bytes);
  stock->handler       = handler;
  stock->connect_flags = connect_flags;
  strcpy(stock->signal, signal);
  return stock;
}

static void
gtracer_connect_log_stock_free(GTracerConnectStock * stock)
{
  g_return_if_fail (stock);
  g_free(stock);
}

static GHashTable *
gtracer_instance_get_stock_table(gpointer instance,
				 gboolean create_if_no_table)
{
  GHashTable * stock_table;
  
  g_return_val_if_fail (instance, NULL);
  stock_table = g_object_get_data(instance, 
				  GTRACER_CONNECT_STOCK_KEY);
  if ((stock_table == NULL) && create_if_no_table)
    {
      stock_table = g_hash_table_new_full(g_direct_hash, 
					  g_direct_equal,
					  NULL,
					  (GDestroyNotify)gtracer_connect_log_stock_free);
      g_object_set_data_full (instance, 
			      GTRACER_CONNECT_STOCK_KEY,
			      stock_table,
			      (GDestroyNotify)g_hash_table_destroy);
    }
  return stock_table;
}

static GTracerConnectStock *
gtracer_instance_steal_stock(gpointer instance, 
			     gulong handler_id)
{
  GHashTable * stock_table;
  GTracerConnectStock * stock;
  g_return_val_if_fail (instance, NULL);
  stock_table = gtracer_instance_get_stock_table(instance, FALSE);
  
  if (stock_table == NULL) 
    return NULL;

  stock = g_hash_table_lookup(stock_table, 
			      GUINT_TO_POINTER(handler_id));
  
  g_hash_table_steal(stock_table, 
		     GUINT_TO_POINTER(handler_id));
  return stock;
}

static GTracerConnectStock *
gtracer_instance_get_stock(gpointer instance, 
			   gulong handler_id)
{
  GHashTable * stock_table;

  g_return_val_if_fail (instance, NULL);
  stock_table = gtracer_instance_get_stock_table(instance, FALSE);
  
  if (stock_table == NULL) 
    return NULL;
  
  return g_hash_table_lookup(stock_table, 
			     GUINT_TO_POINTER(handler_id));
}


/* ------------------------------------------------------------------
   Disconnect
   ------------------------------------------------------------------ */
gulong
gtracer_disconnect_log_handler_id (GTracerDisconnectLog * log)
{
  g_return_val_if_fail (log, 0);
  return log->handler_id;
}

static GTracerLog*
gtracer_disconnect_log_new  (gpointer instance, gulong handler_id, GSList * backtrace)
{
  GTracerDisconnectLog *disconn_log;
  static GTracerLogProtectedOperations disconnect_props = {
    .dump       = gtracer_disconnect_log_dump,
    .free       = gtracer_disconnect_log_free,
  };
  static GTracerSignalLogOperations disconnect_slops = {
    .extract_signal = gtracer_disconnect_log_extract_signal,
    .create_notes   = gtracer_disconnect_log_create_notes,
    .create_message = gtracer_disconnect_log_create_message,
  };
  g_return_val_if_fail (instance, NULL);
  g_return_val_if_fail (handler_id, NULL);
  if (g_inspector_tag_check_object (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }
  if (GTK_IS_WIDGET(instance)
      && g_inspector_tag_check_widget (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }
    
  disconn_log 	     	  = g_new(GTracerDisconnectLog, 1);
  gtracer_signal_log_init((GTracerSignalLog *)disconn_log, 
			  gtracer_disconnect_log_type, 
			  &disconnect_props,
			  instance,
			  backtrace,
			  &disconnect_slops);
  disconn_log->handler_id = handler_id;
  
  return (GTracerLog *)disconn_log;
}

static void
gtracer_disconnect_log_dump (GTracerLogProtected * log_protected)
{
  GTracerSignalLog * log_signal = (GTracerSignalLog *)log_protected;
  GTracerDisconnectLog * disconn_log = (GTracerDisconnectLog *)log_signal;

  g_return_if_fail (disconn_log);
  g_message ("<%s class=\"%s\" address=\"%p\" handler_id=\"%lu\"/>", 
	     GTRACER_DISCONNECT_LOG_NAME, 
	     gtracer_signal_log_class_name(log_signal),
	     gtracer_signal_log_instance(log_signal),
	     disconn_log->handler_id);
}

static void
gtracer_disconnect_log_free (GTracerLogProtected * log_protected)
{
  GTracerDisconnectLog * disconn_log = (GTracerDisconnectLog *)log_protected;
  gpointer instance;
  gulong handler_id;
  GTracerConnectStock * stock;
  
  g_return_if_fail (disconn_log);
  
  instance   = gtracer_signal_log_instance((GTracerSignalLog *)log_protected);
  handler_id = gtracer_disconnect_log_handler_id(disconn_log);
  stock      = gtracer_instance_steal_stock(instance, handler_id);
  if (stock) 
      gtracer_connect_log_stock_free(stock);

  gtracer_signal_log_fin((GTracerSignalLog *)log_protected);
  g_free(disconn_log);
}

gpointer
gtracer_disconnect_log_handler      (GTracerDisconnectLog * log)
{
  gpointer instance;
  gulong handler_id;
  GTracerConnectStock * stock;
  
  g_return_val_if_fail (log, NULL);

  instance   = gtracer_signal_log_instance((GTracerSignalLog*) log);
  handler_id = gtracer_disconnect_log_handler_id(log);
  stock      = gtracer_instance_get_stock(instance, handler_id);
  if (stock == NULL) 
    return NULL;

  return stock->handler;
}

static gchar*
gtracer_disconnect_log_create_notes (GTracerSignalLog * signal_log)
{
  gpointer instance;
  gulong handler_id;

  GTracerConnectStock * stock;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;
  GError  * err = NULL;

  instance   = gtracer_signal_log_instance(signal_log);
  handler_id = gtracer_disconnect_log_handler_id((GTracerDisconnectLog *)signal_log);
  stock      = gtracer_instance_get_stock(instance, handler_id);
  if (stock == NULL) 
    return g_strdup_printf("{static, no debug information} - #%lu", handler_id);

  handler = stock->handler;
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {				/* TODO */
      g_message("%s", err->message);
      g_clear_error(&err);
      return  g_strdup_printf (" - %p # %lu", 
			       handler, 
			       handler_id);
    }

  if (symbol_info == NULL) 
      return g_strdup_printf (" - %p # %lu", 
			      handler,
			      handler_id);

  if (symbol_info->funcname == NULL) 
    return g_strdup_printf (" - %p # %lu", 
			    handler,
			    handler_id);

  return g_strdup_printf (" - %s() # %lu", 
			  symbol_info->funcname,
			  handler_id);
}

static GTracerConnectStock *
gtracer_disconnect_log_get_stock(GTracerDisconnectLog * disconnect_log)
{
  gpointer instance;
  gulong handler_id;
  
  instance   = gtracer_signal_log_instance((GTracerSignalLog *)disconnect_log);
  handler_id = gtracer_disconnect_log_handler_id(disconnect_log);
  return gtracer_instance_get_stock(instance, handler_id);
}

static gchar*
gtracer_disconnect_log_create_message (GTracerSignalLog * signal_log)
{
  GTracerConnectStock * stock;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;
  gchar * msg;
  GError  * err = NULL;

  stock = gtracer_disconnect_log_get_stock((GTracerDisconnectLog *)signal_log);
  if (stock == NULL) 
    return g_strdup("No extra information");

  handler = stock->handler;
  symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
					      &err);
  if (err) 
    {
      msg = g_strdup_printf("Error[symbol]: %s", err->message);
      g_clear_error(&err);
      return msg;
    }

  return gtracer_pmaps_message_new(symbol_info, GPOINTER_TO_UINT(handler));
}

static const gchar*
gtracer_disconnect_log_extract_signal     (GTracerSignalLog * log)
{
  GTracerConnectStock * stock;

  stock = gtracer_disconnect_log_get_stock((GTracerDisconnectLog *)log);
  if (stock == NULL) 
    return NULL;
  else 
    return stock->signal;
}


/* ------------------------------------------------------------------
   Emit
   ------------------------------------------------------------------ */
gboolean
gtracer_emit_log_after  (GTracerEmitLog * log)
{
  g_return_val_if_fail (log, FALSE);
  return log->after;
}

static GTracerLog*
gtracer_emit_log_new  (gpointer instance, guint signal_id, GQuark quark, gboolean after, GSList * backtrace)
{
  GTracerEmitLog * emit_log;
  static GTracerLogProtectedOperations emit_props = {
    .dump = gtracer_emit_log_dump,
    .free = gtracer_emit_log_free,
  };
  static GTracerSignalLogOperations emit_slops = {
    .extract_signal = gtracer_emit_log_extract_signal,
    .create_notes   = gtracer_emit_log_create_notes,
    .create_message = NULL,
  };

  g_return_val_if_fail (instance, NULL);
  g_return_val_if_fail (signal_id, NULL);
  /* g_return_val_if_fail (quark, NULL); */
  if (g_inspector_tag_check_object (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }
  if (GTK_IS_WIDGET(instance)
      && g_inspector_tag_check_widget (instance))
    {
      g_slist_free(backtrace);
      return gtracer_nil_log();
    }

  emit_log 	       	 = g_new(GTracerEmitLog, 1);
  gtracer_signal_log_init((GTracerSignalLog *)emit_log,
			  gtracer_emit_log_type,
			  &emit_props,
			  instance,
			  backtrace,
			  &emit_slops);
  emit_log->signal_id  = signal_id;
  emit_log->after      = after;
  return (GTracerLog *)emit_log;

}

static void
gtracer_emit_log_dump (GTracerLogProtected * log_protected)
{
  GTracerSignalLog * log_signal = (GTracerSignalLog *)log_protected;
  GTracerEmitLog * emit_log = (GTracerEmitLog *)log_signal;

  g_return_if_fail (emit_log);
  g_message ("<%s class=\"%s\" address=\"%p\" signal_id=\"%u\" signal_name=\"%s\" after=\"%s\"/>",
	     GTRACER_EMIT_LOG_NAME,
	     gtracer_signal_log_class_name(log_signal),
	     gtracer_signal_log_instance(log_signal),
	     emit_log->signal_id,
	     gtracer_signal_log_signal(log_signal),
	     gtracer_emit_log_after(emit_log)?"after":"before");
}

static void
gtracer_emit_log_free (GTracerLogProtected * log_protected)
{
  GTracerEmitLog * emit_log = (GTracerEmitLog *)log_protected;
  
  gtracer_signal_log_fin((GTracerSignalLog *)log_protected);

  g_free(emit_log);
}

static gchar*
gtracer_emit_log_create_notes (GTracerSignalLog * signal_log)
{
  return NULL;
}

static const gchar*
gtracer_emit_log_extract_signal(GTracerSignalLog * log)
{
  g_return_val_if_fail (log, NULL);
  return g_signal_name(((GTracerEmitLog *)log)->signal_id);
}
/* arch-tag: d3f84c06-532d-4c7a-a1af-f400d6fd83ca
   (do not change this comment) */
