/* source_log.c --- log class for recording something event about GSource

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "source_service.h"
#include "service.h"
#include "plugin.h"
#include "gtracer.h"

static GQuark gtracer_idle_log_type    = (GQuark)0;
static GQuark gtracer_timeout_log_type = (GQuark)0;
static GQuark gtracer_dropsrc_log_type = (GQuark)0;

struct _GTracerSourceLog
{
  GTracerLogProtected log_protected;

  guint tag;
  gint interval;		/* 0 stands for idle. */
  GSourceFunc function;
  GSource* source;
  GSList * backtrace;
};

static GTracerSourceLog* gtracer_source_log_new  (GQuark log_type,
						  GTracerLogProtectedOperations* props,
						  gint interval,
						  GSourceFunc function,
						  GSource* source,
						  GSList * bt);

static void              gtracer_source_log_free (GTracerLogProtected * log_protected);

static GTracerLog* gtracer_idle_log_new    (GSourceFunc function,
					    GSource* source,
					    GSList * bt);
static void        gtracer_idle_log_dump   (GTracerLogProtected * log_protected);
  
static GTracerLog* gtracer_timeout_log_new  (gint interval,
					     GSourceFunc function,
					     GSource* source,
					     GSList * bt);
static void        gtracer_timeout_log_dump (GTracerLogProtected * log_protected);

struct _GTracerDropsrcLog
{
  GTracerLogProtected log_protected;
  guint tag;
};

static GTracerLog *gtracer_dropsrc_log_new  (guint tag);
static void        gtracer_dropsrc_log_free (GTracerLogProtected * log_protected);
static void        gtracer_dropsrc_log_dump (GTracerLogProtected * log_protected);

/* ------------------------------------------------------------------
   Source Log
   ------------------------------------------------------------------ */
void
gtracer_source_log_class_init (GError ** err)
{
  gtracer_idle_log_type    = g_quark_from_static_string(GTRACER_IDLE_LOG_NAME);
  gtracer_register_log_constructor(gtracer_idle_log_type,
				   gtracer_idle_log_new);
  
  gtracer_timeout_log_type = g_quark_from_static_string(GTRACER_TIMEOUT_LOG_NAME);
  gtracer_register_log_constructor(gtracer_timeout_log_type,
				   gtracer_timeout_log_new);

  gtracer_dropsrc_log_type = g_quark_from_static_string(GTRACER_DROPSRC_LOG_NAME);
  gtracer_register_log_constructor(gtracer_dropsrc_log_type,
				   gtracer_dropsrc_log_new);
  
}

static GTracerSourceLog*
gtracer_source_log_new  (GQuark log_type,
			 GTracerLogProtectedOperations* props,
			 gint interval,
			 GSourceFunc function,
			 GSource* source,
			 GSList * bt)
{
  GTracerSourceLog * source_log;

  g_return_val_if_fail (props, NULL);
  g_return_val_if_fail (function, NULL);
  g_return_val_if_fail (source, NULL);

  source_log = g_new (GTracerSourceLog, 1);
  gtracer_log_init((GTracerLogProtected*) source_log, log_type, props);
  source_log->interval 	= interval;
  source_log->function 	= function;
  source_log->source   	= g_source_ref(source);
  source_log->tag      	= g_source_get_id(source_log->source);
  source_log->backtrace = g_slist_reverse(bt);
  return source_log;
}

static void
gtracer_source_log_free (GTracerLogProtected * log_protected)
{
  gtracer_log_fin(log_protected);
  g_source_unref(((GTracerSourceLog *)log_protected)->source);
  ((GTracerSourceLog *)log_protected)->source = NULL;
  g_slist_free(((GTracerSourceLog *)log_protected)->backtrace);
  ((GTracerSourceLog *)log_protected)->backtrace = NULL;

  g_free(log_protected);
}

static GTracerLog*
gtracer_idle_log_new    (GSourceFunc function,
			 GSource* source,
			 GSList * bt)
{
  static GTracerLogProtectedOperations idle_props = {
    .dump       = gtracer_idle_log_dump,
    .free       = gtracer_source_log_free,
  };
  
  g_return_val_if_fail (function, NULL);
  g_return_val_if_fail (source, NULL);

  return (GTracerLog *)gtracer_source_log_new (gtracer_idle_log_type,
					       &idle_props,
					       0,
					       function,
					       source,
					       bt);
}

static void
gtracer_idle_log_dump   (GTracerLogProtected * log_protected)
{
  GTracerSourceLog * log 	      = (GTracerSourceLog *)log_protected;
  g_message ("<idle>%u</idle>", log->tag);
}

static GTracerLog*
gtracer_timeout_log_new (gint interval,
			 GSourceFunc function,
			 GSource* source,
			 GSList * bt)
{
  static GTracerLogProtectedOperations timeout_props = {
    .dump       = gtracer_timeout_log_dump,
    .free       = gtracer_source_log_free,
  };
  
  g_return_val_if_fail (function, NULL);
  g_return_val_if_fail (source, NULL);

  return (GTracerLog *)gtracer_source_log_new (gtracer_timeout_log_type,
					       &timeout_props,
					       interval,
					       function,
					       source,
					       bt);
}

static void
gtracer_timeout_log_dump (GTracerLogProtected * log_protected)
{
  GTracerSourceLog * log 	      = (GTracerSourceLog *)log_protected;
  g_message ("<timeout>%u</timeout>", log->tag);
}

guint
gtracer_source_log_tag       (GTracerSourceLog * log)
{
  g_return_val_if_fail (log, 0);
  return log->tag;
}

gint
gtracer_source_log_priority  (GTracerSourceLog * log)
{
  GSource * source;
  g_return_val_if_fail (log, 0);
  
  source = log->source;
  return g_source_get_priority(source);
}

guint
gtracer_source_log_interval  (GTracerSourceLog * log)
{
  g_return_val_if_fail (log, 0);
  return log->interval;
}

gpointer
gtracer_source_log_function  (GTracerSourceLog * log)
{
  g_return_val_if_fail (log, NULL);
  return log->function;
}

GSList*
gtracer_source_log_backtrace (GTracerSourceLog * log)
{
  g_return_val_if_fail (log, NULL);
  return log->backtrace;
}
/* ------------------------------------------------------------------
   Dropsrc Log
   ------------------------------------------------------------------ */
static GTracerLog *
gtracer_dropsrc_log_new  (guint tag)
{
  GTracerDropsrcLog * dropsrc_log;
  static GTracerLogProtectedOperations dropsrc_props = {
    .dump       = gtracer_dropsrc_log_dump,
    .free       = gtracer_dropsrc_log_free,
  };

  dropsrc_log = g_new (GTracerDropsrcLog, 1);
  gtracer_log_init((GTracerLogProtected *)dropsrc_log, 
		   gtracer_dropsrc_log_type, 
		   &dropsrc_props);
  dropsrc_log->tag = tag;
  return (GTracerLog *)dropsrc_log;
}

static void
gtracer_dropsrc_log_free (GTracerLogProtected * log_protected)
{
  GTracerDropsrcLog * dropsrc_log = (GTracerDropsrcLog *)log_protected;
  g_return_if_fail (dropsrc_log);
  gtracer_log_fin(log_protected);
  g_free(log_protected);
}

static void
gtracer_dropsrc_log_dump (GTracerLogProtected * log_protected)
{
  GTracerDropsrcLog * dropsrc_log = (GTracerDropsrcLog *)log_protected;
  g_message("<dropsrc>%u</dropsrc>", dropsrc_log->tag);
}

guint
gtracer_dropsrc_log_tag      (GTracerDropsrcLog * log)
{
  g_return_val_if_fail (log, 0);
  return log->tag;
}

/* arch-tag: 2d060648-697f-4182-b781-119494b746cc
   (do not change this comment) */
