/* usignal_service.c --- implementation of unix signal manipulator

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include <signal.h>
#include "usignal_service.h"
#include "service.h"

#include "i18n.h"

#include <ginspector/ginspector-tag.h>

enum
{
  COL_NUMBER = 0,
  COL_HANDLER = 1,
  COL_SYMBOLINFO = 2,
  NUM_COLS
} ;

void
gtracer_usignal_service_init (GError** err)
{
  gtracer_usignal_viewer_init(err);
}
  
GtkTreeModel*
gtracer_usignal_model_new (GError ** err)
{
  GtkListStore  *store;
  GError * tmp_error = NULL;
  
  store = gtk_list_store_new (NUM_COLS, G_TYPE_INT, G_TYPE_POINTER, AddrslvSymbolInfoType);
  g_inspector_tag_to_object(G_OBJECT(store));

  gtracer_usignal_model_update(GTK_TREE_MODEL(store), &tmp_error);
  if (tmp_error)
    {
      g_propagate_error(err, tmp_error);
      g_object_unref(G_OBJECT(store));
      return NULL;
    } 
  return GTK_TREE_MODEL(store);
}

void
gtracer_usignal_model_update(GtkTreeModel* model, GError ** err)
{
  GtkTreeIter iter;
  gint i;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;

  gtk_list_store_clear(GTK_LIST_STORE(model));

  for (i = 0; i < 0xFF; i++)
    {
      handler = signal(i, SIG_DFL);
      signal(i, handler);
      
      if (handler == SIG_ERR) 
	continue ;

      if (handler == SIG_DFL || handler == SIG_IGN) 
	symbol_info = NULL;
      else
	symbol_info = gtracer_pmaps_address_resolve(GPOINTER_TO_UINT(handler),
						    NULL);      

      gtk_list_store_append (GTK_LIST_STORE(model), &iter);
      gtk_list_store_set (GTK_LIST_STORE(model), &iter, COL_NUMBER, i, -1);
      gtk_list_store_set (GTK_LIST_STORE(model), &iter, COL_HANDLER, handler, -1);
      gtk_list_store_set (GTK_LIST_STORE(model), &iter, COL_SYMBOLINFO, symbol_info, -1);

    }
}

void
gtracer_usignal_model_extract_number (GtkTreeViewColumn *col,
				      GtkCellRenderer   *renderer,
				      GtkTreeModel      *model,
				      GtkTreeIter       *iter,
				      gpointer           user_data)
{
  gint n;
  gchar buffer[64];
  gtk_tree_model_get(model, iter, COL_NUMBER, &n, -1);
  snprintf(buffer, 64, "%d", n);
  g_object_set(renderer, "text", buffer, NULL);
}

void
gtracer_usignal_model_extract_name (GtkTreeViewColumn *col,
				    GtkCellRenderer   *renderer,
				    GtkTreeModel      *model,
				    GtkTreeIter       *iter,
				    gpointer           user_data)
{
  gint n;
  const gchar * name;
  
  gtk_tree_model_get(model, iter, COL_NUMBER, &n, -1);
  name = g_strsignal(n);
  if (name)
    g_object_set(renderer, "text", name, NULL);
  else
    g_object_set(renderer, 
		 "text", _("{static, no debug information}"), 
		 NULL);
}

void
gtracer_usignal_model_extract_action (GtkTreeViewColumn *col,
				      GtkCellRenderer   *renderer,
				      GtkTreeModel      *model,
				      GtkTreeIter       *iter,
				      gpointer           user_data)
{
  AddrslvSymbolInfo * info = NULL;
  gpointer handler 	   = NULL;
  
  gtk_tree_model_get(model, iter, COL_HANDLER, &handler, -1);
  gtk_tree_model_get(model, iter, COL_SYMBOLINFO, &info, -1);
  if (info && info->funcname)
    g_object_set(renderer, "text", info->funcname, NULL);
  else if (handler == SIG_DFL)
    g_object_set(renderer, "text", _("{default action}"), NULL);
  else if (handler == SIG_IGN)
    g_object_set(renderer, "text", _("{ignoring}"), NULL);
  else
    {
      gchar buffer[64];
      snprintf(buffer, 64, "%8p", handler);
      g_object_set(renderer, "text", buffer, NULL);
    }

  if (info) 
    g_boxed_free(AddrslvSymbolInfoType, info);/* MODEL_GET_FREE */
}

gchar*
gtracer_usignal_model_extract_message (GtkTreeModel * model,
				       GtkTreeIter *iter)
{
  gchar * msg;
  gpointer handler;
  AddrslvSymbolInfo * symbol_info;

  gtk_tree_model_get (model, iter, COL_HANDLER, &handler, -1);
  gtk_tree_model_get (model, iter, COL_SYMBOLINFO, &symbol_info, -1);
  
  if (handler == NULL) 
    return NULL;

  msg = gtracer_pmaps_message_new(symbol_info, 
				  GPOINTER_TO_UINT(handler));
  if (symbol_info) 
    g_boxed_free(AddrslvSymbolInfoType, symbol_info);
  return msg;
}

/* arch-tag: 4cf52180-e61d-4a26-9898-1969e1d342bd
   (do not change this comment) */
