/* maps_loader.c --- loading memory mapping from /proc/$pid/maps

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "loader.h"

#if defined HAVE_MAPS

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

static gboolean maps_init         (PMaps* pmaps, GError** err);
static gboolean maps_load         (PMaps* pmaps, GError** err);
static void     maps_fin          (PMaps* pmaps);

static guint  maps_parse_protection (gchar * prot_string);

static PMapsLoader loader = {
  .init         = maps_init,
  .load         = maps_load,
  .fin          = maps_fin,
};
PMapsLoader* pmaps_maps_loader = &loader;

static gboolean
maps_init (PMaps* pmaps, GError** err)
{
  FILE*  maps_fp;
  gchar* maps_filename;
  pid_t pid;

  g_return_val_if_fail (pmaps, 1);
  pid = pmaps_get_pid(pmaps);
  
  maps_filename = g_strdup_printf("/proc/%u/maps", pid);

  errno = 0;
  maps_fp = fopen(maps_filename, "r");

  if (maps_fp == NULL)
    g_set_error(err,
		PMAPS_ERRNO_ERROR,
		errno,
		"%s: %s", maps_filename, g_strerror (errno));

  g_free(maps_filename);
  
  pmaps_set_loader_data(pmaps, maps_fp);
  return  (maps_fp)? TRUE: FALSE;
}

static gboolean
maps_load       (PMaps* pmaps, GError** err)
{
  FILE* maps_fp;
  gchar buffer[PATH_MAX];
  
  g_return_val_if_fail (pmaps, 1);

  maps_fp = pmaps_get_loader_data(pmaps);
  g_return_val_if_fail (maps_fp, 2);
  rewind(maps_fp);

  errno = 0;
  while(fgets(buffer, PATH_MAX - 1, maps_fp))
    {
      guint start, end;
      gchar prot_string[5];
      guint prot;
      gchar filename[PATH_MAX];
  
      gint  scan_count;
      gchar unused[32];

      filename[0] = '\0';
      scan_count = sscanf(buffer,
			  "%x-%x "
			  "%4s "
			  "%32s %32s %32s "
			  "%s",
			  &start, &end,
			  prot_string, 
			  unused, unused, unused,
			  filename);
      if (scan_count == 7
	  || ((scan_count    == 6) && (filename[0] == '\0')))
	{
	  prot = maps_parse_protection(prot_string);
	  pmaps_add_entry(pmaps, start, end, prot, filename);
	}
    }

  if (errno)
    {
      g_set_error(err,
		  PMAPS_ERRNO_ERROR,
		  errno,
		  "%s", g_strerror (errno));
      return FALSE;
    }
  else
    return TRUE;
}

static guint
maps_parse_protection (gchar * prot_string)
{
  guint prot = 0;
  g_return_val_if_fail (strlen(prot_string) == 4, 0);
  
  if (prot_string[0] == 'r')
    prot |= PMAPS_READABLE;
  
  if (prot_string[1] == 'w')
    prot |= PMAPS_WRITABLE;

  if (prot_string[2] == 'x')
    prot |= PMAPS_EXECUTABLE;

  if (prot_string[3] != 'p')
    prot |= PMAPS_SHARED;
  return prot;
}

static void
maps_fin          (PMaps* pmaps)
{
  FILE * maps_fp;

  g_return_if_fail (pmaps);
  maps_fp = pmaps_get_loader_data(pmaps);
  if (maps_fp)
    fclose(maps_fp);
  pmaps_set_loader_data(pmaps, NULL);
}
#else
PMapsLoader* pmaps_maps_loader = NULL;
#endif /* Def: HAVE_MAPS */

/* arch-tag: c941e262-ec57-440a-8ead-15e1a7d2a0f0
   (do not change this comment) */
