/* pmap_loader.c --- loading memory mapping from pmap command

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "loader.h"

#if defined HAVE_PMAP

#include <stdio.h>
#include <string.h>
#include <unistd.h>

static gboolean pmap_init         (PMaps* pmaps, GError** err);
static gboolean pmap_load         (PMaps* pmaps, GError** err);
static void     pmap_fin          (PMaps* pmaps);

static guint  pmap_parse_protection       (gchar * prot_string);

static PMapsLoader loader = {
  .init         = pmap_init,
  .load         = pmap_load,
  .fin          = pmap_fin,
};
PMapsLoader* pmaps_pmap_loader = &loader;

static gboolean
pmap_init (PMaps* pmaps, GError** err)
{
  gchar* pid_string;
  pid_t pid;
  
  g_return_val_if_fail (pmaps, 1);
  pid = pmaps_get_pid(pmaps);

  pid_string = g_strdup_printf("%lu", (gulong)pid);
  
  pmaps_set_loader_data(pmaps, pid_string);
  return TRUE;
}

static gboolean
pmap_load       (PMaps* pmaps, GError** err)
{
  gchar*   pmap_argv[3];
  gchar*   pmap_rawdata = NULL;
  gboolean pmap_spawn_return;

  gchar** pmap_lines;
  gchar** buffer;

  GError * tmp_error;
  
  pmap_argv[0] 	    = PMAP_PATH;
  pmap_argv[1] 	    = pmaps_get_loader_data(pmaps);
  pmap_argv[2] 	    = NULL;
  
  pmap_spawn_return = g_spawn_sync("/", 
				   pmap_argv, NULL,
				   G_SPAWN_STDERR_TO_DEV_NULL, 
				   /* setup, user_data */
				   NULL, NULL, 
				   /* stdout, stderr */
				   &pmap_rawdata, NULL, 
				   /* status */
				   NULL,	     
				   &tmp_error);
  
  if (pmap_spawn_return == FALSE) {
    g_propagate_error (err, tmp_error);
    return FALSE;
  }

  if (pmap_rawdata[0] == '\0'  ) {
    /* Empty??? */
    g_free(pmap_rawdata);
    return TRUE;		
  }
    
  pmap_lines = g_strsplit(pmap_rawdata, "\n", 0);
  for (buffer = pmap_lines; *buffer != NULL; buffer++)
    {
      guint start, size;
      gchar prot_string[32];
      guint prot;
      gchar filename[PATH_MAX];
      gint  scan_count;

      scan_count = sscanf(*buffer,
			  "%x %uK %32s %s",
			  &start, &size,
			  prot_string, 
			  filename);

      if (scan_count == 4 
	  && filename
	  /* FIXME: Ignore stack and heap now */
	  && (filename[0] != '['))
	{
	  prot = pmap_parse_protection(prot_string);
	  pmaps_add_entry(pmaps, 
			  start, 
			  start + 1024 * size, 
			  prot, 
			  filename);
	}
      /* TODO: The case when the filename is empty is not supported yet. */
    }

  g_strfreev(pmap_lines);
  g_free(pmap_rawdata);

  return TRUE;
}

static guint
pmap_parse_protection (gchar * prot_string)
{
  guint prot = 0;
  g_return_val_if_fail (prot_string, 0);

  /* NetBSD */
  if (strstr(prot_string, "read"))
    prot |= PMAPS_READABLE;

  if (strstr(prot_string, "write"))
    prot |= PMAPS_WRITABLE;

  if (strstr(prot_string, "exec"))
    prot |= PMAPS_EXECUTABLE;

  if (strlen(prot_string) != 5)
    return prot;
  
  /* Solaris 9 */
  if (prot_string[0] == 'r')
    prot |= PMAPS_READABLE;
  
  if (prot_string[1] == 'w')
    prot |= PMAPS_WRITABLE;

  if (prot_string[2] == 'x')
    prot |= PMAPS_EXECUTABLE;
  
  return prot;
}

static void
pmap_fin          (PMaps* pmaps)
{
  gchar* pid_string;

  g_return_if_fail (pmaps);
  pid_string = pmaps_get_loader_data(pmaps);
  g_free(pid_string);
  pmaps_set_loader_data(pmaps, NULL);
}
#else
PMapsLoader* pmaps_pmap_loader = NULL;
#endif /* Def: HAVE_PMAP */

/* arch-tag: dece4516-47b2-4e69-848d-a263b5c9d72b
   (do not change this comment) */
