/* pmaps.c --- Process memory mapping manipulator

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "pmaps.h"
#include "loader.h"

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif  /* G_LOG_DOMAIN */
#define PMAPS_LOG_DOMAIN "LibPMaps"
#define G_LOG_DOMAIN PMAPS_LOG_DOMAIN

typedef struct _PMapsUpdateMarhallData PMapsUpdateMarhallData;

struct _PMaps 
{
  pid_t   pid;
  GSList* list;
  gpointer loader_data;

  GData*         attached_data;

  GHookList update_hook_list;

  guint ref_count;
};

struct _PMapsUpdateMarhallData
{
  PMaps*   pmaps;
  gboolean after;
};

/* PMaps manipulators */
static void        pmaps_reset         (PMaps* pmaps);
static void        pmaps_free          (PMaps* pmaps);

/* PMapsEntry manipulators */
static PMapsEntry* pmaps_entry_new     (guint start, guint end, guint prot, gchar * filename);
static void        pmaps_entry_free    (PMapsEntry * entry);
static gint        pmaps_entry_compare (gconstpointer a, gconstpointer b);

/* Container --- entries managements */
static void        pmaps_container_init    (PMaps * pmaps);
static void        pmaps_container_fin     (PMaps * pmaps);
static void        pmaps_container_add     (PMaps * pmaps, PMapsEntry * entry);
static PMapsEntry* pmaps_container_find    (PMaps * pmaps, guint address);
static void        pmaps_container_foreach (PMaps* pmaps, PMapsForeachFunc func, gpointer callback_data);

/* PMaps update notify marshaller */
static void        pmaps_update_notify_marshaller(GHook		*hook,
						  gpointer	 marshal_data);

/* Loader */
static PMapsLoader* pmaps_loader = NULL;

/* 
 * PMaps manipulators
 */
PMaps*
pmaps_new     (pid_t pid, GError** err)
{
  PMaps* pmaps;
  GError* tmp_error = NULL;
  
  pmaps_loader =
#if defined HAVE_MAPS
    pmaps_maps_loader
#elif defined HAVE_PMAP
    pmaps_pmap_loader
#else
#   error "no pmaps loader is available for this platform."
#endif /* defined */
    ;
  pmaps = g_new(PMaps, 1);
  pmaps->ref_count    = 1;
  
  pmaps->pid 	      = pid;
  pmaps_container_init(pmaps);

  g_datalist_init(&pmaps->attached_data);
  g_hook_list_init(&pmaps->update_hook_list,
		   sizeof(GHook));
  pmaps->loader_data = NULL;
  if (pmaps_loader->init(pmaps, &tmp_error) == FALSE)
    {
      g_propagate_error (err, tmp_error);
      pmaps_unref(pmaps);
      return NULL;
    }

  if (pmaps_update (pmaps, &tmp_error) == FALSE)
    {
      g_propagate_error (err, tmp_error);
      pmaps_unref(pmaps);
      return NULL;
    }

  return pmaps; 
}

PMaps*
pmaps_ref     (PMaps* pmaps)
{
  g_return_val_if_fail (pmaps, NULL);
  g_return_val_if_fail (pmaps->ref_count > 0, NULL);

  pmaps->ref_count++;
  return pmaps;
}

void
pmaps_unref   (PMaps* pmaps)
{
  g_return_if_fail (pmaps);
  g_return_if_fail (pmaps->ref_count > 0);
  pmaps->ref_count--;  
  
  if (pmaps->ref_count == 0)
    pmaps_free(pmaps);
}

static void
pmaps_reset(PMaps* pmaps)
{
  g_return_if_fail (pmaps);
  pmaps_container_fin(pmaps);
}

static void
pmaps_free (PMaps* pmaps)
{
  g_return_if_fail (pmaps);

  g_datalist_clear(&pmaps->attached_data);
  g_hook_list_clear(&pmaps->update_hook_list);
  
  pmaps_reset(pmaps);

  pmaps_loader->fin(pmaps);
  
  g_free(pmaps);
}
  
gboolean
pmaps_update(PMaps* pmaps, GError** err)
{
  gboolean update_result;
  GError* tmp_error = NULL;
  PMapsUpdateMarhallData marshal_data;
  
  g_return_val_if_fail (pmaps, 1);

  marshal_data.pmaps = pmaps;
  marshal_data.after = FALSE;
  g_hook_list_marshal(&pmaps->update_hook_list,
		      TRUE,
		      pmaps_update_notify_marshaller,
		      &marshal_data);

  pmaps_reset(pmaps);
  update_result = pmaps_loader->load(pmaps, &tmp_error);
  if (update_result == FALSE)
    g_propagate_error (err, tmp_error);

  marshal_data.after = TRUE;
  g_hook_list_marshal(&pmaps->update_hook_list,
		      TRUE,
		      pmaps_update_notify_marshaller,
		      &marshal_data);
  
  return update_result;
}

PMapsEntry *
pmaps_find(PMaps* pmaps, guint address)
{
  return pmaps_container_find(pmaps, address);
}

void
pmaps_foreach (PMaps* pmaps, PMapsForeachFunc func, gpointer callback_data)
{
  pmaps_container_foreach(pmaps, func, callback_data);
}

void
pmaps_set_data (PMaps* pmaps,
		const gchar* key,
		gpointer data,
		GDestroyNotify data_destroy_func)
{
  g_return_if_fail (pmaps);
  g_return_if_fail (key);

  g_datalist_set_data_full(&pmaps->attached_data, 
			   key, 
			   data, 
			   data_destroy_func);
}

gpointer
pmaps_get_data (PMaps* pmaps, const gchar* key)
{
  g_return_val_if_fail (pmaps, NULL);
  g_return_val_if_fail (key, NULL);

  return g_datalist_get_data(&pmaps->attached_data, key);
}

gulong
pmaps_add_update_notify    (PMaps* pmaps, 
			    PMapsUpdateNotify update_notify_func,
			    gpointer callback_data)
{
  GHook * hook;

  g_return_val_if_fail (pmaps, (gulong)-1);
  g_return_val_if_fail (update_notify_func, (gulong)-1);
  
  hook 	     = g_hook_alloc(&pmaps->update_hook_list);
  hook->data = callback_data;
  hook->func = update_notify_func;
  g_hook_prepend(&pmaps->update_hook_list, hook);
  return hook->hook_id;
}

gboolean
pmaps_remove_update_notify (PMaps* pmaps, 
			    gulong notify_id)
{
  GHook * hook;
  
  g_return_val_if_fail (pmaps, FALSE);
  
  hook = g_hook_get(&pmaps->update_hook_list,
		    notify_id);
  if (hook)
    {
      /* ??? */
      g_hook_destroy_link(&pmaps->update_hook_list,
			  hook); 
      return TRUE;
    }
  else
    return FALSE;
}


pid_t
pmaps_get_pid          (PMaps * pmaps)
{
  g_return_val_if_fail (pmaps, 0);
  return pmaps->pid;
}

gpointer
pmaps_get_loader_data  (PMaps * pmaps)
{
  g_return_val_if_fail (pmaps, NULL);
  return pmaps->loader_data;
}

void
pmaps_set_loader_data  (PMaps * pmaps, gpointer loader_data)
{
  g_return_if_fail (pmaps);
  pmaps->loader_data = loader_data;
}

void
pmaps_add_entry(PMaps * pmaps,
		guint start, guint end, guint prot, gchar * filename)
{
  PMapsEntry * entry;
  g_return_if_fail (pmaps);
  
  entry = pmaps_entry_new(start, end, prot, filename);
  pmaps_container_add(pmaps, entry);
}

void
pmaps_dump (PMaps * pmaps, const gchar* prefix)
{
  guint i = 0;
  i++;
  
  if (prefix == NULL)
    prefix = "";
  
  if (pmaps == NULL)
    {
      g_message("[%03u: %8s] pmaps = null", i, prefix);
      return ;
    }

  g_message("[%03u: %8s] pid: %u, ref_count: %u ", i, prefix, 
	    pmaps->pid, pmaps->ref_count);

  pmaps_foreach(pmaps, 
		(PMapsForeachFunc)pmaps_entry_dump, 
		(gpointer)prefix);

}
/* 
 * PMapsEntry manipulators 
 */

PMapsEntry*
pmaps_entry_ref      (PMapsEntry* entry)
{
  g_return_val_if_fail (entry, NULL);
  g_return_val_if_fail (entry->ref_count > 0, NULL);
  entry->ref_count++;
  return entry;
}

void
pmaps_entry_unref    (PMapsEntry* entry)
{
  g_return_if_fail (entry);
  g_return_if_fail (entry->ref_count > 0);
  entry->ref_count--;
  if (entry->ref_count == 0)
    pmaps_entry_free(entry);
}

gint
pmaps_entry_include (const PMapsEntry* entry, guint address)
{
  if (entry->end <= address)
    return -1;
  else if (entry->start > address)
    return 1;
  else
    return 0;
}


void
pmaps_entry_set_data(PMapsEntry* entry,
		     const gchar* key,
		     gpointer data, 
		     GDestroyNotify data_destroy_func)
{
  g_return_if_fail (entry);
  g_return_if_fail (key);

  g_datalist_set_data_full(&entry->attached_data, key, data, data_destroy_func);
}
  
gpointer
pmaps_entry_get_data (PMapsEntry* entry, const gchar* key)
{
  g_return_val_if_fail (entry, NULL);
  g_return_val_if_fail (key, NULL);

  return g_datalist_get_data(&entry->attached_data, key);
}

static PMapsEntry *
pmaps_entry_new (guint start, guint end, guint prot, gchar * filename)
{
  PMapsEntry * entry = g_new(PMapsEntry, 1);
  
  entry->start 	       		= start;
  entry->end   	       		= end;
  entry->prot  	       		= prot;
  entry->filename      		= g_strdup(filename);
  entry->ref_count 		= 1;
  
  g_datalist_init(&entry->attached_data);

  return entry;
}

static void
pmaps_entry_free(PMapsEntry * entry)
{
  g_datalist_clear(&entry->attached_data);
  g_free(entry->filename);
  g_free(entry);
}


static gint
pmaps_entry_compare (gconstpointer a, gconstpointer b)
{
  const PMapsEntry * entry = a;
  guint address     = GPOINTER_TO_UINT(b);
  return pmaps_entry_include(entry, address);
}

void
pmaps_entry_dump (PMapsEntry * entry, const gchar* prefix)
{
  static guint i = 0;
  i++;
  
  if (prefix == NULL) 
    prefix = "";
  
  if (entry == NULL)
    {
      g_message("[%03u: entry %8s] pmaps entry: null", i, prefix);
      return ;
    }

  g_message("[%03u: entry %8s] 0x%08x => 0x%08x %c%c%c%c %s, ref_count: %u", 
	    i, prefix, 
	    entry->start, entry->end,
	    (entry->prot & PMAPS_READABLE)?   'r': '-',
	    (entry->prot & PMAPS_WRITABLE)?   'w': '-',
	    (entry->prot & PMAPS_EXECUTABLE)? 'x': '-',
	    (entry->prot & PMAPS_SHARED)?     '-': 'p',
	    entry->filename,
	    entry->ref_count);
}
  
/*
 * Container --- Entries managements 
 * ---------------------------------
 * TODO: Use garray or gtree
 */

static void
pmaps_container_init(PMaps * pmaps)
{
  pmaps->list = NULL;
}

static void
pmaps_container_fin(PMaps * pmaps)
{
  GSList * tmp;
  tmp = pmaps->list;
  while (tmp) 
    {
      pmaps_entry_unref(tmp->data);
      tmp = g_slist_next(tmp);
    }
  g_slist_free(pmaps->list);
  pmaps->list = NULL;

}
static void
pmaps_container_add(PMaps * pmaps, PMapsEntry * entry)
{
  pmaps->list = g_slist_append(pmaps->list, entry);
}

static PMapsEntry *
pmaps_container_find    (PMaps * pmaps, guint address)
{
  GSList * tmp;

  tmp = g_slist_find_custom(pmaps->list, 
			    GUINT_TO_POINTER(address), 
			    pmaps_entry_compare);
  return tmp? tmp->data: NULL;
}

static void
pmaps_container_foreach (PMaps* pmaps, PMapsForeachFunc func, gpointer callback_data)
{
  g_slist_foreach(pmaps->list, (GFunc)func, callback_data);
}

/* 
 * PMaps update notify marshaller 
 */
static void
pmaps_update_notify_marshaller(GHook		*hook,
			       gpointer	 marshal_data)
{
  PMapsUpdateNotify func = hook->func;
  
  g_return_if_fail (marshal_data);

  func(((PMapsUpdateMarhallData*)marshal_data)->pmaps, 
       ((PMapsUpdateMarhallData*)marshal_data)->after,
       hook->data);

}

/* 
 * GError related
 */
GQuark
pmaps_errno_error_quark (void)
{
  static GQuark quark = 0;

  if (!quark)
    quark = g_quark_from_static_string ("pmaps_errno_error");
  
  return quark;
}

/* arch-tag: b0106b13-5350-42d5-9b5c-a873804ff0bd
   (do not change this comment) */
