/* signal_wrapper.c --- Recording when signal functions are invoked

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include "preload.h"

#include <glib.h>
#include <glib-object.h>
#include <stdio.h>
#include <backtrace.h>

/* -----------------------------------------------------------------
   HELPER: Context data strage functions
   -----------------------------------------------------------------
   Push and pop mechanism to track thread context in wrappers. 
   ----------------------------------------------------------------- */ 
static void         gtracer_callback_push_instance(gpointer instance);
static void         gtracer_callback_pop_instance(void);
static gpointer     gtracer_callback_get_instance(void);

static void         gtracer_callback_push_signal(const char * signal);
static void         gtracer_callback_pop_signal(void);
static const char * gtracer_callback_get_signal(void);

static void         gtracer_callback_push_handler(GCallback handler);
static void         gtracer_callback_pop_handler(void);
static GCallback    gtracer_callback_get_handler(void);

/* -----------------------------------------------------------------
   WRAPPER: g_signal_connect_data
   -----------------------------------------------------------------
   o Push and pop signal data to the thread storage which is 
     refereed in `g_cclosure_new'. 
   o Enqueue GTracerConnectLog. 
   ----------------------------------------------------------------- */ 
gulong
g_signal_connect_data(gpointer       instance,
		      const gchar   *detailed_signal,
		      GCallback      c_handler,
		      gpointer       data,
		      GClosureNotify destroy_data,
		      GConnectFlags  connect_flags)
{
  static gulong (* GENUINE) (gpointer       instance,
			     const gchar   *detailed_signal,
			     GCallback      c_handler,
			     gpointer       data,
			     GClosureNotify destroy_data,
			     GConnectFlags  connect_flags) = NULL;
  CUTIN_VAR;
  gulong result;
  static GTracerConnectLogNew connect_log_new = NULL;
  gboolean zombiep;
  GSList * bt = NULL;
  
  g_assert(instance);
  zombiep = gtracer_is_object_zombie(instance);
  
  CUTIN_IF_PRE(!zombiep)
    {
      gtracer_callback_push_instance(instance);
      gtracer_callback_push_handler(c_handler);
      gtracer_callback_push_signal(detailed_signal);
    }

  GENUINE_GET();
  result = GENUINE (instance,
		    detailed_signal,
		    c_handler,
		    data,
		    destroy_data,
		    connect_flags);


  CUTIN_IF_POST(!zombiep)
    {
      if (G_UNLIKELY(connect_log_new == NULL))
	connect_log_new = gtracer_preload_get_log_constructor(GTRACER_CONNECT_LOG_NAME);
      if (G_LIKELY(connect_log_new))
	{
	  GTracerLog * log;

	  bt = backtrace_funcall();
	  log = connect_log_new(instance, detailed_signal, c_handler, 
				connect_flags, result, bt);
	  gtracer_preload_queue_push_log(log);
	  gtracer_preload_log_unref(log);
	}
      gtracer_callback_pop_signal();
      gtracer_callback_pop_handler();
      gtracer_callback_pop_instance();
    }
  
  return result;
}

/* -----------------------------------------------------------------
   WRAPPER: g_cclosure_new
   -----------------------------------------------------------------
   For installing marshall guard 
   ----------------------------------------------------------------- */ 
static void gtracer_callback_guard (gpointer  data,
				    GClosure *closure);
static void gtracer_callback_finalize_notifier (gpointer  data,
						GClosure *closure);

GClosure* 
g_cclosure_new (GCallback      callback_func,
		gpointer       user_data,
		GClosureNotify destroy_data)
{
  static GClosure* (* GENUINE) (GCallback      callback_func,
				gpointer       user_data,
				GClosureNotify destroy_data) = NULL;
  CUTIN_VAR;
  GClosure* result;
  static GTracerCallbackLogNew callback_log_new = NULL;

  CUTIN_IF_PRE(TRUE) 
    {
      /* Do nothing */
    }
  
  GENUINE_GET();
  result = GENUINE ( callback_func, user_data, destroy_data );

  CUTIN_IF_PRE(TRUE)
    {
      /* Check whther this `g_cclosure_new' invocation is called
	 in `g_signal_connect_data' or not. If not, this invocation
	 is nothing to do with `signal', just a closure. */
      if ( result && callback_func != gtracer_callback_get_handler() )
	return result;
  
      if (G_UNLIKELY(callback_log_new == NULL))
	callback_log_new = gtracer_preload_get_log_constructor(GTRACER_CALLBACK_LOG_NAME);

      if (G_UNLIKELY(callback_log_new == NULL))
	return result;
  
      {
	GTracerLog * log_before;
	GTracerLog * log_after;
	gpointer instance    = gtracer_callback_get_instance();
	const char * signal  = gtracer_callback_get_signal();
	GCallback  handler   = gtracer_callback_get_handler();
      
	log_before = callback_log_new(instance, signal, handler, FALSE);
	log_after  = callback_log_new(instance, signal, handler, TRUE);
      
	g_closure_add_marshal_guards(result,
				     log_before,
				     gtracer_callback_guard,
				     log_after,
				     gtracer_callback_guard);
	g_closure_add_finalize_notifier(result,
					log_before,
					gtracer_callback_finalize_notifier);
	g_closure_add_finalize_notifier(result,
					log_after,
					gtracer_callback_finalize_notifier);
      }
    }
  return result;
}


static void
gtracer_callback_guard(gpointer	 data,
		       GClosure	*closure)
{
  static GTracerProxyLogNew proxy_log_new = NULL;
  GTracerLog * proxy_log;
  
  g_return_if_fail (data);

  if (G_UNLIKELY(proxy_log_new == NULL))
    proxy_log_new = gtracer_preload_get_log_constructor(GTRACER_PROXY_LOG_NAME);
  g_assert(proxy_log_new);
      
  proxy_log = proxy_log_new(data);
  gtracer_preload_queue_push_log(proxy_log);
  gtracer_preload_log_unref(proxy_log);
}

static void
gtracer_callback_finalize_notifier(gpointer  data,
				   GClosure *closure)
{
  GTracerLog * log = data;
  gtracer_preload_log_unref(log);
}

/* -----------------------------------------------------------------
   WRAPPER: g_signal_handler_disconnect
   ----------------------------------------------------------------- */ 
void
g_signal_handler_disconnect(gpointer		  instance,
			    gulong		  handler_id)
{
  static void (* GENUINE) (gpointer		  instance,
			   gulong		  handler_id);
  CUTIN_VAR;
  
  GTracerLog * log;
  static GTracerDisconnectLogNew disconnect_log_new = NULL;
  gboolean zombiep;
  GSList * bt = NULL;
  
  g_assert(instance);
  zombiep = gtracer_is_object_zombie(instance);
  
  CUTIN_IF_PRE(!zombiep)
    {
      /* Do nothing */
    }
  GENUINE_GET();
  GENUINE(instance, handler_id);

  CUTIN_IF_POST(!zombiep)
    {
      if (G_UNLIKELY(disconnect_log_new == NULL))
	disconnect_log_new = gtracer_preload_get_log_constructor(GTRACER_DISCONNECT_LOG_NAME);
      if (G_UNLIKELY(disconnect_log_new == NULL))
	return ;
  
      bt = backtrace_funcall();
      log = disconnect_log_new(instance, handler_id, bt);
      gtracer_preload_queue_push_log(log);
      gtracer_preload_log_unref(log);
    }
}

/* -----------------------------------------------------------------
   WRAPPER: g_signal_emit_valist
   ----------------------------------------------------------------- */ 
void
g_signal_emit_valist (gpointer instance,
		      guint    signal_id,
		      GQuark   detail,
		      va_list  var_args)
{
  static void (* GENUINE) (gpointer instance,
			   guint    signal_id,
			   GQuark   detail,
			   va_list  var_args);  
  CUTIN_VAR;
  
  GTracerLog * log;
  static GTracerEmitLogNew emit_log_new = NULL;
  gboolean zombiep;
  GSList * bt = NULL;
  
  g_assert(instance);
  zombiep = gtracer_is_object_zombie(instance);
  
  CUTIN_IF_PRE(!zombiep)
    {
      if (G_UNLIKELY(emit_log_new == NULL))
	emit_log_new = gtracer_preload_get_log_constructor(GTRACER_EMIT_LOG_NAME);
      if (G_LIKELY(emit_log_new != NULL))
	{
	  bt = backtrace_funcall();
	  log = emit_log_new(instance, signal_id, detail, FALSE, bt);
	  gtracer_preload_queue_push_log(log);
	  gtracer_preload_log_unref(log);
	}
    }

  GENUINE_GET();
  GENUINE(instance, signal_id, detail, var_args);

  CUTIN_IF_POST(!zombiep)
    {
      if (G_LIKELY(emit_log_new != NULL))
	{
	  bt = backtrace_funcall();
	  log = emit_log_new(instance, signal_id, detail, TRUE, bt);
	  gtracer_preload_queue_push_log(log);
	  gtracer_preload_log_unref(log);
	}
    }
}

/* -----------------------------------------------------------------
   WRAPPER: g_signal_emitv
   ----------------------------------------------------------------- */ 
void
g_signal_emitv (const GValue *instance_and_params,
		guint         signal_id,
		GQuark	      detail,
		GValue       *return_value)
{
  static void (* GENUINE) (const GValue *instance_and_params,
			   guint         signal_id,
			   GQuark	 detail,
			   GValue       *return_value);
  CUTIN_VAR;
  
  GTracerLog * log;
  static GTracerEmitLogNew emit_log_new = NULL;
  gpointer instance 			= NULL;
  gboolean zombiep;
  GSList * bt;
  
  g_assert(instance_and_params);
  instance = g_value_peek_pointer (instance_and_params);
  g_assert(instance);
  zombiep = gtracer_is_object_zombie(instance);
  
  CUTIN_IF_PRE(!zombiep)
    {
      if (G_UNLIKELY(emit_log_new == NULL))
	emit_log_new = gtracer_preload_get_log_constructor(GTRACER_EMIT_LOG_NAME);
      
      if (G_LIKELY(emit_log_new != NULL))
	{
	  bt = backtrace_funcall();
	  log = emit_log_new(instance, signal_id, detail, FALSE, bt);
	  gtracer_preload_queue_push_log(log);
	  gtracer_preload_log_unref(log);
	}
    }
  
  GENUINE_GET();    
  GENUINE (instance_and_params,
	   signal_id,
	   detail,
	   return_value);

  CUTIN_IF_POST(!zombiep)
    {
      if (G_LIKELY(emit_log_new != NULL))
	{
	  bt = backtrace_funcall();
	  log = emit_log_new(instance, signal_id, detail, TRUE, bt);
	  gtracer_preload_queue_push_log(log);
	  gtracer_preload_log_unref(log);
	}
    }
}

/* ---------------------------------------------------------------------
   HELPER: Context data strage functions
   ---------------------------------------------------------------------
   Push and pop mechanism to track thread context in wrappers. 
   -------------------------------------------------------------------- */
#ifdef G_THREADS_ENABLED
static GStaticPrivate gtracer_callback_signal_private   = G_STATIC_PRIVATE_INIT;
static GStaticPrivate gtracer_callback_handler_private  = G_STATIC_PRIVATE_INIT;
static GStaticPrivate gtracer_callback_instance_private = G_STATIC_PRIVATE_INIT;
#endif /* Def: G_THREADS_ENABLED */

/* 
 * signal 
 */
static void
gtracer_callback_push_signal(const char * signal)
{
  g_return_if_fail (NULL == gtracer_callback_get_signal());
  g_static_private_set (&gtracer_callback_signal_private, (gpointer)signal, NULL);
}

static void
gtracer_callback_pop_signal(void)
{
  g_static_private_set (&gtracer_callback_signal_private, NULL, NULL);
}

static const char *
gtracer_callback_get_signal(void)
{
  return g_static_private_get (&gtracer_callback_signal_private);
}

/* 
 * handler
 */
static void
gtracer_callback_push_handler(GCallback handler)
{
  g_return_if_fail (NULL == gtracer_callback_get_handler());
  g_static_private_set (&gtracer_callback_handler_private, 
			handler, 
			NULL);
}

static void
gtracer_callback_pop_handler(void)
{
  g_static_private_set (&gtracer_callback_handler_private, 
			NULL, 
			NULL);
}

static GCallback
gtracer_callback_get_handler(void)
{
  return g_static_private_get (&gtracer_callback_handler_private);
}

/* 
 * instance
 */
static void
gtracer_callback_push_instance(gpointer instance)
{
  g_return_if_fail (NULL == gtracer_callback_get_instance());
  g_static_private_set (&gtracer_callback_instance_private, 
			instance, 
			NULL);
}

static void
gtracer_callback_pop_instance(void)
{
  g_static_private_set (&gtracer_callback_instance_private, 
			NULL, 
			NULL);
}

static gpointer
gtracer_callback_get_instance(void)
{
  return g_static_private_get (&gtracer_callback_instance_private);
}

/* arch-tag: 1cc1dc9c-6e29-4415-a581-6a8311455919
   (do not change this comment) */
