/* source_wrapper.c --- Recording when GSource functions are invoked

   Copyright (c) 2004, 2005 by Masatake YAMATO and Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.  

   - Neither the name of Red Hat, Inc nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
   OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Development of the code in this file is support of 
   Information-technology Promotion Agency(IPA), Japan. */

#include <glib.h>
#include "preload.h"
#include "functor.h"
#include <backtrace.h>

static void gtracer_dropsrc_empty_destroy_notify         (gpointer data) {;}
static void gtracer_dropsrc_idle_destroy_notify_after    (guint tag, gpointer functor);
static void gtracer_dropsrc_timeout_destroy_notify_after (guint tag, gpointer functor);

guint
g_idle_add_full(gint priority,
		GSourceFunc func,
		gpointer data,
		GDestroyNotify notify)
{
  static guint (* GENUINE)(gint priority,
			   GSourceFunc func,
			   gpointer data,
			   GDestroyNotify notify) = NULL;
  CUTIN_VAR;
  guint tag;
  static GTracerIdleLogNew idle_log_new = NULL;
  GSource * source;
  GDestroyNotify wrapper;
  

  if (notify == NULL) 
    notify  = gtracer_dropsrc_empty_destroy_notify;
  wrapper = gtracer_functor_new(notify,
				gtracer_dropsrc_idle_destroy_notify_after);
  
  GENUINE_GET();
  tag = GENUINE(priority, func, data, wrapper);
  source = g_main_context_find_source_by_id (NULL, tag);
  gtracer_functor_set_tag(wrapper, tag);
  
  CUTIN_IF_POST(source != NULL)
    {
      GTracerLog * log;
      GSList * bt;
      
      if (G_UNLIKELY(idle_log_new == NULL))
	idle_log_new = gtracer_preload_get_log_constructor(GTRACER_IDLE_LOG_NAME);
      g_assert(idle_log_new);
      
      bt = backtrace_funcall();
      log = idle_log_new(func, source, bt);
      gtracer_preload_queue_push_log(log);
      gtracer_preload_log_unref(log);
    }

  return tag;
}

guint
g_timeout_add_full              (gint priority,
				 guint interval,
				 GSourceFunc function,
				 gpointer data,
				 GDestroyNotify notify)
{
  static guint (* GENUINE) (gint priority,
			    guint interval,
			    GSourceFunc function,
			    gpointer data,
			    GDestroyNotify notify) = NULL;
  CUTIN_VAR;
  guint tag;
  static GTracerTimeoutLogNew timeout_log_new = NULL;
  GSource * source;
  GDestroyNotify wrapper;
  
  if (notify == NULL) 
    notify = gtracer_dropsrc_empty_destroy_notify;
  wrapper  = gtracer_functor_new(notify,
				 gtracer_dropsrc_timeout_destroy_notify_after);
  
  GENUINE_GET();
  tag = GENUINE(priority, interval, function, data, wrapper);
  source = g_main_context_find_source_by_id (NULL, tag);
  gtracer_functor_set_tag(wrapper, tag);
  
  CUTIN_IF_POST(source != NULL)
    {
      GTracerLog * log;
      GSList * bt;

      if (G_UNLIKELY(timeout_log_new == NULL))
	timeout_log_new = gtracer_preload_get_log_constructor(GTRACER_TIMEOUT_LOG_NAME);
      g_assert(timeout_log_new);
      
      bt = backtrace_funcall();
      log = timeout_log_new(interval, function, source, bt);
      gtracer_preload_queue_push_log(log);
      gtracer_preload_log_unref(log);
    }
  
  return tag;
}

static void
gtracer_dropsrc_idle_destroy_notify_after    (guint tag, gpointer functor)
{
  static GTracerDropsrcLogNew dropsrc_log_new = NULL;
  CUTIN_VAR;
  CUTIN
    {
      GTracerLog * log;
      
      if (G_UNLIKELY(dropsrc_log_new == NULL))
	dropsrc_log_new = gtracer_preload_get_log_constructor(GTRACER_DROPSRC_LOG_NAME);
      g_assert(dropsrc_log_new);

      log = dropsrc_log_new(tag);
      gtracer_preload_queue_push_log(log);
      gtracer_preload_log_unref(log);
    }
  gtracer_functor_free(functor);
}

static void
gtracer_dropsrc_timeout_destroy_notify_after (guint tag, gpointer functor)
{
  static GTracerDropsrcLogNew dropsrc_log_new = NULL;
  CUTIN_VAR;
  CUTIN
    {
      GTracerLog * log;
      
      if (G_UNLIKELY(dropsrc_log_new == NULL))
	dropsrc_log_new = gtracer_preload_get_log_constructor(GTRACER_DROPSRC_LOG_NAME);
      g_assert(dropsrc_log_new);

      log = dropsrc_log_new(tag);
      gtracer_preload_queue_push_log(log);
      gtracer_preload_log_unref(log);
    }
  gtracer_functor_free(functor);
}

/* arch-tag: 17d9b7fd-116d-46e5-872d-890ff6458088
   (do not change this comment) */
